#  SPDX-License-Identifier: Apache-2.0
"""
Python Package for controlling Tesla API.

For more details about this api, please refer to the documentation at
https://github.com/zabuldon/teslajsonpy
"""
import logging
from typing import Callable

from teslajsonpy.const import DEFAULT_ENERGYSITE_NAME, RESOURCE_TYPE

_LOGGER = logging.getLogger(__name__)


class EnergySite:
    """Base class to represents a Tesla energy site."""

    def __init__(self, api: Callable, energysite: dict, site_config: dict) -> None:
        """Initialize EnergySite."""
        self._api = api
        self._energysite = energysite
        self._site_config = site_config

    @property
    def energysite_id(self) -> int:
        """Return energy site id (aka site_id)."""
        return self._energysite.get("energy_site_id")

    @property
    def has_load_meter(self) -> bool:
        """Return True if energy site has a load meter."""
        return self._energysite.get("components", {}).get("load_meter")

    @property
    def has_battery(self) -> bool:
        """Return True if energy site has battery."""
        return self._energysite.get("components", {}).get("battery")

    @property
    def has_solar(self) -> bool:
        """Return True if energy site has solar."""
        return self._energysite.get("components", {}).get("solar")

    @property
    def id(self) -> str:
        # pylint: disable=invalid-name
        """Return battery_id."""
        return self._energysite.get("id")

    @property
    def resource_type(self) -> str:
        """Return energy site type."""
        return self._energysite[RESOURCE_TYPE]

    async def _send_command(
        self, name: str, *, path_vars: dict, wake_if_asleep: bool = False, **kwargs
    ) -> dict:
        """Wrap commands sent to Tesla API."""
        _LOGGER.debug("Sending command: %s", name)
        data = await self._api(
            name, path_vars=path_vars, wake_if_asleep=wake_if_asleep, **kwargs
        )
        _LOGGER.debug("Response from command %s: %s", name, data)
        return data


class SolarSite(EnergySite):
    """Represents a Tesla Energy Solar site.

    This class shouldn't be instantiated directly; it will be instantiated
    by :meth:`teslajsonpy.controller.generate_energysite_objects`.
    """

    def __init__(
        self, api: Callable, energysite: dict, site_config: dict, site_data: dict
    ) -> None:
        """Initialize SolarSite."""
        super().__init__(api, energysite, site_config)
        self._site_data = site_data

    @property
    def data_available(self) -> bool:
        """Return if data is available."""
        return self._site_data != {}

    @property
    def grid_power(self) -> float:
        """Return grid power in Watts."""
        return self._site_data.get("grid_power")

    @property
    def load_power(self) -> float:
        """Return load power in Watts."""
        return self._site_data.get("load_power")

    @property
    def site_name(self) -> str:
        """Return energy site name."""
        # "site_name" not a valid key if name never set in Tesla app
        return self._site_config.get("site_name", DEFAULT_ENERGYSITE_NAME)

    @property
    def solar_power(self) -> float:
        """Return solar power in Watts."""
        return self._site_data.get("solar_power")

    @property
    def solar_type(self) -> str:
        """Return type of solar (e.g. pv_panels or roof)."""
        return self._energysite.get("components", {}).get("solar_type")


class PowerwallSite(EnergySite):
    """Represents a Tesla Energy Powerwall site.

    This class shouldn't be instantiated directly; it will be instantiated
    by :meth:`teslajsonpy.controller.generate_energysite_objects`.
    """

    def __init__(
        self,
        api: Callable,
        energysite: dict,
        site_config: dict,
        site_data: dict,
        site_summary: dict,
    ) -> None:
        """Initialize PowerwallSite."""
        super().__init__(api, energysite, site_config)
        self._site_data: dict = site_data
        self._site_summary: dict = site_summary

    @property
    def backup_reserve_percent(self) -> int:
        """Return backup reserve percentage."""
        return self._site_config.get("backup_reserve_percent")

    @property
    def battery_power(self) -> float:
        """Return battery power in Watts."""
        return self._site_summary.get("battery_power")

    @property
    def data_available(self) -> bool:
        """Return if data is available."""
        return self._site_summary != {}

    @property
    def energy_left(self) -> float:
        """Return battery energy left in Watt hours."""
        return round (self._site_config.get("nameplate_energy") * self.percentage_charged / 100 )

    @property
    def grid_power(self) -> float:
        """Return grid power in Watts."""
        return self._site_data.get("grid_power")

    @property
    def grid_status(self) -> str:
        """Return grid status."""
        return self._site_data.get("grid_status")

    @property
    def load_power(self) -> float:
        """Return load power in Watts."""
        return self._site_data.get("load_power")

    @property
    def operation_mode(self) -> str:
        """Return operation mode."""
        return self._site_config.get("default_real_mode")

    @property
    def percentage_charged(self) -> float:
        """Return battery percentage charged."""
        # percentage_charged sometimes incorrectly reports 0
        return self._site_summary.get("percentage_charged")

    @property
    def site_name(self) -> str:
        """Return energy site name."""
        # "site_name" not a valid key if name never set in Tesla app
        return self._site_summary.get("site_name", DEFAULT_ENERGYSITE_NAME)

    @property
    def solar_power(self) -> float:
        """Return solar power in Watts."""
        return self._site_data.get("solar_power")

    @property
    def version(self) -> float:
        """Return firmware version."""
        return self._site_config.get("version")

    async def set_operation_mode(self, real_mode: str) -> None:
        """Set operation mode of Powerwall.

        Mode: "self_consumption", "backup", "autonomous"
        """
        data = await self._send_command(
            "BATTERY_OPERATION_MODE",
            path_vars={"site_id": self.energysite_id},
            default_real_mode=real_mode,
        )
        if data and data["response"]["code"] == 201:
            self._site_config.update({"operation": real_mode})

    async def set_reserve_percent(self, value: int) -> None:
        """Set reserve percentage of Powerwall.

        Value: 0-100
        """
        data = await self._send_command(
            "BACKUP_RESERVE",
            path_vars={"site_id": self.energysite_id},
            backup_reserve_percent=int(value),
        )
        if data and data["response"]["code"] == 201:
            self._site_config.update({"backup_reserve_percent": value})


class SolarPowerwallSite(PowerwallSite):
    """Represents a Tesla Energy Solar site with Powerwall(s).

    This class shouldn't be instantiated directly; it will be instantiated
    by :meth:`teslajsonpy.controller.generate_energysite_objects`.
    """

    @property
    def export_rule(self) -> str:
        """Return energy export rule setting."""
        return self._site_config.get("components", {}).get(
            "customer_preferred_export_rule"
        )

    @property
    def grid_charging(self) -> bool:
        """Return grid charging."""
        # Key is missing from site config when False
        return not self._site_config.get("components", {}).get(
            "disallow_charge_from_grid_with_solar_installed", False
        )

    @property
    def solar_type(self) -> str:
        """Return type of solar (e.g. pv_panels or roof)."""
        return self._site_config.get("components", {}).get("solar_type")

    async def set_grid_charging(self, value: bool) -> None:
        """Set grid charging setting of Powerwall."""
        value = not value
        await self._send_command(
            "ENERGY_SITE_IMPORT_EXPORT_CONFIG",
            path_vars={"site_id": self.energysite_id},
            disallow_charge_from_grid_with_solar_installed=value,
        )
        # This endpoint returns an empty response instead of a result code
        self._site_config["components"].update(
            {"disallow_charge_from_grid_with_solar_installed": value}
        )

    async def set_export_rule(self, setting: str) -> None:
        """Set energy export setting of Powerwall.

        Settings
          Solar: "pv_only"
          Everything: "battery_ok"
        """
        await self._send_command(
            "ENERGY_SITE_IMPORT_EXPORT_CONFIG",
            path_vars={"site_id": self.energysite_id},
            customer_preferred_export_rule=setting,
        )
        # This endpoint returns an empty response instead of a result code
        self._site_config["components"].update(
            {"customer_preferred_export_rule": setting}
        )



#  SPDX-License-Identifier: Apache-2.0
"""
Python Package for controlling Tesla API.

For more details about this api, please refer to the documentation at
https://github.com/zabuldon/teslajsonpy
"""
import logging
from typing import Optional, Tuple

from teslajsonpy.exceptions import HomelinkError, TeslaException

_LOGGER = logging.getLogger(__name__)

CLIMATE_KEEPER_ID_MAP = {
    0: "off",
    1: "on",
    2: "dog",
    3: "camp",
}

SEAT_ID_MAP = {
    0: "left",
    1: "right",
    2: "rear_left",
    4: "rear_center",
    5: "rear_right",
    6: "third_row_left",
    7: "third_row_right",
}

DAY_SELECTION_MAP = {
    "all_week": False,
    "weekdays": True,
}


class TeslaCar:
    #  pylint: disable=too-many-public-methods
    """Represents a Tesla car.

    This class shouldn't be instantiated directly; it will be instantiated
    by :meth:`teslajsonpy.controller.generate_car_objects`.
    """

    def __init__(self, car: dict, controller, vehicle_data: dict) -> None:
        """Initialize TeslaCar."""
        self._car = car
        self._controller = controller
        self._vehicle_data = vehicle_data

        self._previous_driver_temp = self.driver_temp_setting
        self._previous_fan_status = self.fan_status
        self._previous_passenger_temp = self.passenger_temp_setting

    @property
    def display_name(self) -> str:
        """Return display name."""
        return self._car.get("display_name")

    @property
    def id(self) -> int:
        # pylint: disable=invalid-name
        """Return id."""
        return self._car.get("id")

    @property
    def state(self) -> str:
        """Return car state."""
        return self._car.get("state")

    @property
    def vehicle_id(self) -> int:
        """Return car id."""
        return self._car.get("vehicle_id")

    @property
    def vin(self) -> str:
        """Return car vin."""
        return self._car.get("vin")

    @property
    def data_available(self) -> bool:
        """Return if data from VEHICLE_DATA endpoint is available."""
        # self._vehicle_data gets updated with some data from VEHICLE_LIST endpoint
        # Only return True if data specifically from VEHICLE_DATA endpoint is available
        # vehicle_state is only from the VEHICLE_DATA endpoint
        if self._vehicle_data.get("vehicle_state", {}):
            return True
        return None

    @property
    def battery_level(self) -> float:
        """Return car battery level (SOC). This is not affected by temperature."""
        return self._vehicle_data.get("charge_state", {}).get("battery_level")

    @property
    def usable_battery_level(self) -> float:
        """Return car usable battery level (uSOE). This is the value used in the app and car."""
        return self._vehicle_data.get("charge_state", {}).get("usable_battery_level")

    @property
    def battery_range(self) -> float:
        """Return car battery range."""
        return self._vehicle_data.get("charge_state", {}).get("battery_range")

    @property
    def cabin_overheat_protection(self) -> str:
        """Return cabin overheat protection."""
        return self._vehicle_data.get("climate_state", {}).get(
            "cabin_overheat_protection"
        )

    @property
    def car_type(self) -> str:
        """Return car type."""
        return f"Model {str(self.vin[3]).upper()}"

    @property
    def car_version(self) -> str:
        """Return installed car software version."""
        return self._vehicle_data.get("vehicle_state", {}).get("car_version")

    @property
    def charger_actual_current(self) -> int:
        """Return charger actual current."""
        return self._vehicle_data.get("charge_state", {}).get("charger_actual_current")

    @property
    def charge_current_request(self) -> int:
        """Return charge current request."""
        return self._vehicle_data.get("charge_state", {}).get("charge_current_request")

    @property
    def charge_current_request_max(self) -> int:
        """Return charge current request max."""
        return self._vehicle_data.get("charge_state", {}).get(
            "charge_current_request_max"
        )

    @property
    def charge_port_latch(self) -> str:
        """Return charger port latch state.

        Returns
            str: Engaged

        Other states?

        """
        return self._vehicle_data.get("charge_state", {}).get("charge_port_latch")

    @property
    def charge_energy_added(self) -> float:
        """Return charge energy added."""
        return self._vehicle_data.get("charge_state", {}).get("charge_energy_added")

    @property
    def charge_limit_soc(self) -> int:
        """Return charge limit soc."""
        return self._vehicle_data.get("charge_state", {}).get("charge_limit_soc")

    @property
    def charge_limit_soc_max(self) -> int:
        """Return charge limit soc max."""
        return self._vehicle_data.get("charge_state", {}).get("charge_limit_soc_max")

    @property
    def charge_limit_soc_min(self) -> int:
        """Return charge limit soc min."""
        return self._vehicle_data.get("charge_state", {}).get("charge_limit_soc_min")

    @property
    def charge_miles_added_ideal(self) -> float:
        """Return charge ideal miles added."""
        return self._vehicle_data.get("charge_state", {}).get(
            "charge_miles_added_ideal"
        )

    @property
    def charge_miles_added_rated(self) -> float:
        """Return charge rated miles added."""
        return self._vehicle_data.get("charge_state", {}).get(
            "charge_miles_added_rated"
        )

    @property
    def charger_phases(self) -> int:
        """Return charger phase."""
        return self._vehicle_data.get("charge_state", {}).get("charger_phases")

    @property
    def charger_power(self) -> int:
        """Return charger power."""
        return self._vehicle_data.get("charge_state", {}).get("charger_power")

    @property
    def charge_rate(self) -> str:
        """Return charge rate."""
        return self._vehicle_data.get("charge_state", {}).get("charge_rate")

    @property
    def charging_state(self) -> str:
        """Return charging state.

        Returns
            str: Charging, Stopped, Complete, Disconnected, NoPower
            None: When car is asleep

        """
        return self._vehicle_data.get("charge_state", {}).get("charging_state")

    @property
    def charger_voltage(self) -> int:
        """Return charger voltage."""
        return self._vehicle_data.get("charge_state", {}).get("charger_voltage")

    @property
    def bioweapon_mode(self) -> bool:
        """Return bioweapon defense mode."""
        return self._vehicle_data.get("climate_state", {}).get("bioweapon_mode")
    
    @property
    def climate_keeper_mode(self) -> str:
        """Return climate keeper mode mode.

        Returns
            str: dog, camp, on, off

        Not supported on all Tesla models.

        """
        return self._vehicle_data.get("climate_state", {}).get("climate_keeper_mode")

    @property
    def conn_charge_cable(self) -> str:
        """Return charge cable connection."""
        return self._vehicle_data.get("charge_state", {}).get("conn_charge_cable")

    @property
    def defrost_mode(self) -> int:
        """Return defrost mode.

        Returns
            int: 2 (on), 0 (off)

        """
        return self._vehicle_data.get("climate_state", {}).get("defrost_mode", 0)

    @property
    def driver_temp_setting(self) -> float:
        """Return driver temperature setting."""
        return self._vehicle_data.get("climate_state", {}).get("driver_temp_setting")

    @property
    def fan_status(self) -> int:
        """Return fan status setting."""
        return self._vehicle_data.get("climate_state", {}).get("fan_status")

    @property
    def fast_charger_present(self) -> bool:
        """Return fast charger present."""
        return self._vehicle_data.get("charge_state", {}).get("fast_charger_present")

    @property
    def fast_charger_brand(self) -> str:
        """Return fast charger brand."""
        return self._vehicle_data.get("charge_state", {}).get("fast_charger_brand")

    @property
    def fast_charger_type(self) -> str:
        """Return fast charger type."""
        return self._vehicle_data.get("charge_state", {}).get("fast_charger_type")

    @property
    def door_df(self) -> int:
        """Return driver front door status."""
        return self._vehicle_data.get("vehicle_state", {}).get("df")

    @property
    def door_dr(self) -> int:
        """Return driver rear door status."""
        return self._vehicle_data.get("vehicle_state", {}).get("dr")

    @property
    def door_pf(self) -> int:
        """Return passenger front door status."""
        return self._vehicle_data.get("vehicle_state", {}).get("pf")

    @property
    def door_pr(self) -> int:
        """Return passenger rear door status."""
        return self._vehicle_data.get("vehicle_state", {}).get("pr")

    @property
    def gui_distance_units(self) -> str:
        """Return gui distance units."""
        return self._vehicle_data.get("gui_settings", {}).get("gui_distance_units")

    @property
    def gui_range_display(self) -> str:
        """Return range display."""
        return self._vehicle_data.get("gui_settings", {}).get("gui_range_display")

    @property
    def heading(self) -> int:
        """Return heading."""
        return self._vehicle_data.get("drive_state", {}).get("heading")

    @property
    def homelink_device_count(self) -> int:
        """Return Homelink device count."""
        return self._vehicle_data.get("vehicle_state", {}).get("homelink_device_count")

    @property
    def homelink_nearby(self) -> bool:
        """Return Homelink nearby."""
        return self._vehicle_data.get("vehicle_state", {}).get("homelink_nearby")

    @property
    def ideal_battery_range(self) -> float:
        """Return car ideal battery range."""
        return self._vehicle_data.get("charge_state", {}).get("ideal_battery_range")

    @property
    def in_service(self) -> bool:
        """Return car in_service."""
        if self.data_available:
            return self._vehicle_data.get("in_service")
        return None

    @property
    def inside_temp(self) -> float:
        """Return inside temperature."""
        return self._vehicle_data.get("climate_state", {}).get("inside_temp")

    @property
    def is_charge_port_door_open(self) -> bool:
        """Return charger port door open."""
        return self._vehicle_data.get("charge_state", {}).get("charge_port_door_open")

    @property
    def is_climate_on(self) -> bool:
        """Return climate is on."""
        return self._vehicle_data.get("climate_state", {}).get("is_climate_on")

    @property
    def is_frunk_closed(self) -> bool:
        """Return car frunk is closed.

        Returns
            bool: True (0), False (255)

        """
        response = self._vehicle_data.get("vehicle_state", {}).get("ft")
        return response == 0

    @property
    def is_in_gear(self) -> bool:
        """Return car is gear (i.e. drive or reverse)."""
        return self.shift_state in ["D", "R"]

    @property
    def is_locked(self) -> bool:
        """Return car is locked."""
        return self._vehicle_data.get("vehicle_state", {}).get("locked")

    @property
    def is_steering_wheel_heater_on(self) -> bool:
        """Return steering wheel heater."""
        return self._vehicle_data.get("climate_state", {}).get("steering_wheel_heater")

    @property
    def is_trunk_closed(self) -> bool:
        """Return car trunk is closed.

        Returns
            bool: True (0), False (1-255)

        """
        response = self._vehicle_data.get("vehicle_state", {}).get("rt")
        return response == 0

    @property
    def is_on(self) -> bool:
        """Return car is on."""
        return self._controller.is_car_online(vin=self.vin)

    @property
    def longitude(self) -> float:
        """Return longitude."""
        return self._vehicle_data.get("drive_state", {}).get("longitude")

    @property
    def latitude(self) -> float:
        """Return latitude."""
        return self._vehicle_data.get("drive_state", {}).get("latitude")

    @property
    def max_avail_temp(self) -> float:
        """Return max available temperature."""
        return self._vehicle_data.get("climate_state", {}).get("max_avail_temp")

    @property
    def min_avail_temp(self) -> float:
        """Return min available temperature."""
        return self._vehicle_data.get("climate_state", {}).get("min_avail_temp")

    @property
    def native_heading(self) -> int:
        """Return native heading."""
        return self._vehicle_data.get("drive_state", {}).get("native_heading")

    @property
    def native_location_supported(self) -> int:
        """Return native location supported."""
        return self._vehicle_data.get("drive_state", {}).get(
            "native_location_supported"
        )

    @property
    def native_longitude(self) -> float:
        """Return native longitude."""
        return self._vehicle_data.get("drive_state", {}).get("native_longitude")

    @property
    def native_latitude(self) -> float:
        """Return native latitude."""
        return self._vehicle_data.get("drive_state", {}).get("native_latitude")

    @property
    def native_type(self) -> float:
        """Return native type."""
        return self._vehicle_data.get("drive_state", {}).get("native_type")

    @property
    def odometer(self) -> float:
        """Return odometer."""
        return self._vehicle_data.get("vehicle_state", {}).get("odometer")

    @property
    def outside_temp(self) -> float:
        """Return outside temperature."""
        return self._vehicle_data.get("climate_state", {}).get("outside_temp")

    @property
    def passenger_temp_setting(self) -> float:
        """Return passenger temperature setting."""
        return self._vehicle_data.get("climate_state", {}).get("passenger_temp_setting")

    @property
    def power(self) -> int:
        """Return power."""
        return self._vehicle_data.get("drive_state", {}).get("power")

    @property
    def powered_lift_gate(self) -> bool:
        """Return True if car has power lift gate."""
        return self._vehicle_data.get("vehicle_config", {}).get("plg")

    @property
    def rear_seat_heaters(self) -> int:
        """Return if car has rear (second row) heated seats.

        Returns
            int: 0 (no rear heated seats), int: ? (rear heated seats)

        """
        return self._vehicle_data.get("vehicle_config", {}).get("rear_seat_heaters")

    @property
    def has_seat_cooling(self) -> bool:
        """Return if car has cooled seats."""
        return self._vehicle_data.get("vehicle_config", {}).get("has_seat_cooling")

    @property
    def sentry_mode(self) -> bool:
        """Return sentry mode."""
        return self._vehicle_data.get("vehicle_state", {}).get("sentry_mode")

    @property
    def sentry_mode_available(self) -> bool:
        """Return sentry mode available."""
        return self._vehicle_data.get("vehicle_state", {}).get("sentry_mode_available")

    @property
    def shift_state(self) -> str:
        """Return shift state."""
        return self._vehicle_data.get("drive_state", {}).get("shift_state")

    @property
    def speed(self) -> float:
        """Return speed."""
        return self._vehicle_data.get("drive_state", {}).get("speed")

    @property
    def software_update(self) -> dict:
        """Return software update version information."""
        return self._vehicle_data.get("vehicle_state", {}).get("software_update")

    @property
    def steering_wheel_heater(self) -> bool:
        """Return steering wheel heater option."""
        return (
            self._vehicle_data.get("climate_state", {}).get("steering_wheel_heater")
            is not None
        )

    @property
    def pedestrian_speaker(self) -> Optional[bool]:
        """Return pedestrian warning speaker option."""
        if self._vehicle_data.get("option_codes", {}) is None:
            return None
        if "P3WS" in self._vehicle_data.get("option_codes", {}):
            return True
        return False

    @property
    def tpms_pressure_fl(self) -> float:
        """Return tire pressure sensor for front left tire."""
        return self._vehicle_data.get("vehicle_state", {}).get("tpms_pressure_fl")

    @property
    def tpms_pressure_fr(self) -> float:
        """Return tire pressure sensor for front right tire."""
        return self._vehicle_data.get("vehicle_state", {}).get("tpms_pressure_fr")

    @property
    def tpms_pressure_rl(self) -> float:
        """Return tire pressure sensor for rear left tire."""
        return self._vehicle_data.get("vehicle_state", {}).get("tpms_pressure_rl")

    @property
    def tpms_pressure_rr(self) -> float:
        """Return tire pressure sensor for rear right tire."""
        return self._vehicle_data.get("vehicle_state", {}).get("tpms_pressure_rr")

    @property
    def third_row_seats(self) -> str:
        """Return third row seats option.

        Returns
            str: None

        """
        return self._vehicle_data.get("vehicle_config", {}).get("third_row_seats")

    @property
    def time_to_full_charge(self) -> float:
        """Return time to full charge."""
        return self._vehicle_data.get("charge_state", {}).get("time_to_full_charge")

    @property
    def window_fd(self) -> int:
        """Return front driver window status."""
        return self._vehicle_data.get("vehicle_state", {}).get("fd_window")

    @property
    def window_fp(self) -> int:
        """Return front passenger window status."""
        return self._vehicle_data.get("vehicle_state", {}).get("fp_window")

    @property
    def window_rd(self) -> int:
        """Return rear driver window status."""
        return self._vehicle_data.get("vehicle_state", {}).get("rd_window")

    @property
    def window_rp(self) -> int:
        """Return rear passenger window status."""
        return self._vehicle_data.get("vehicle_state", {}).get("rp_window")

    @property
    def is_window_closed(self) -> bool:
        """Return all car windows are close."""
        if (
            self._vehicle_data.get("vehicle_state", {}).get("fd_window")
            or self._vehicle_data.get("vehicle_state", {}).get("fp_window")
            or self._vehicle_data.get("vehicle_state", {}).get("rd_window")
            or self._vehicle_data.get("vehicle_state", {}).get("rp_window")
        ):
            return False
        return True

    @property
    def is_remote_start(self) -> bool:
        """Return if remote start active."""
        return self._vehicle_data.get("vehicle_state", {}).get("remote_start")

    @property
    def is_valet_mode(self) -> bool:
        """Return state of valet mode."""
        return self._vehicle_data.get("vehicle_state", {}).get("valet_mode")

    @property
    def is_auto_seat_climate_left(self) -> bool:
        """Return state of auto seat climate left."""
        return self._vehicle_data.get("climate_state", {}).get("auto_seat_climate_left")

    @property
    def is_auto_seat_climate_right(self) -> bool:
        """Return state of auto seat climate right."""
        return self._vehicle_data.get("climate_state", {}).get(
            "auto_seat_climate_right"
        )

    @property
    def is_auto_steering_wheel_heat(self) -> bool:
        """Return the state of auto steering wheel heat."""
        return self._vehicle_data.get("climate_state", {}).get(
            "auto_steering_wheel_heat"
        )

    @property
    def active_route_destination(self) -> Optional[str]:
        """Return active route destination."""
        if "active_route_destination" in self._vehicle_data.get("drive_state", {}):
            return self._vehicle_data.get("drive_state", {}).get(
                "active_route_destination"
            )
        return None

    @property
    def active_route_energy_at_arrival(self) -> Optional[int]:
        """Return active route latitude."""
        if "active_route_energy_at_arrival" in self._vehicle_data.get(
            "drive_state", {}
        ):
            return self._vehicle_data.get("drive_state", {}).get(
                "active_route_energy_at_arrival"
            )
        return None

    @property
    def active_route_latitude(self) -> Optional[float]:
        """Return active route latitude."""
        if "active_route_latitude" in self._vehicle_data.get("drive_state", {}):
            return self._vehicle_data.get("drive_state", {}).get(
                "active_route_latitude"
            )
        return None

    @property
    def active_route_longitude(self) -> Optional[float]:
        """Return active route longitude."""
        if "active_route_longitude" in self._vehicle_data.get("drive_state", {}):
            return self._vehicle_data.get("drive_state", {}).get(
                "active_route_longitude"
            )
        return None

    @property
    def active_route_miles_to_arrival(self) -> Optional[float]:
        """Return active route latitude."""
        if "active_route_miles_to_arrival" in self._vehicle_data.get("drive_state", {}):
            return self._vehicle_data.get("drive_state", {}).get(
                "active_route_miles_to_arrival"
            )
        return None

    @property
    def active_route_minutes_to_arrival(self) -> Optional[float]:
        """Return active route minutes to arrival."""
        if "active_route_minutes_to_arrival" in self._vehicle_data.get(
            "drive_state", {}
        ):
            return self._vehicle_data.get("drive_state", {}).get(
                "active_route_minutes_to_arrival"
            )
        return None

    @property
    def active_route_traffic_minutes_delay(self) -> Optional[float]:
        """Return active route traffic minutes delay."""
        if "active_route_minutes_to_arrival" in self._vehicle_data.get(
            "drive_state", {}
        ):
            return self._vehicle_data.get("drive_state", {}).get(
                "active_route_traffic_minutes_delay"
            )
        return None

    @property
    def scheduled_departure_time(self) -> int:
        """Return the scheduled departure time."""
        return self._vehicle_data.get("charge_state", {}).get(
            "scheduled_departure_time"
        )

    @property
    def scheduled_departure_time_minutes(self) -> int:
        """Return the scheduled departure time in minutes after midnight."""
        return self._vehicle_data.get("charge_state", {}).get(
            "scheduled_departure_time_minutes"
        )

    @property
    def is_off_peak_charging_enabled(self) -> bool:
        """Return if peak charging is enabled for scheduled departure."""
        return self._vehicle_data.get("charge_state", {}).get(
            "off_peak_charging_enabled"
        )

    @property
    def is_off_peak_charging_weekday_only(self) -> bool:
        """Return if off off peak charging is weekday only for scheduled departure."""
        return DAY_SELECTION_MAP.get(
            self._vehicle_data.get("charge_state", {}).get("off_peak_charging_times")
        )

    @property
    def off_peak_hours_end_time(self) -> int:
        """Return end of off peak hours in minutes after midnight for scheduled departure."""
        return self._vehicle_data.get("charge_state", {}).get("off_peak_hours_end_time")

    @property
    def is_preconditioning_enabled(self) -> bool:
        """Return if preconditioning is enabled for scheduled departure."""
        return self._vehicle_data.get("charge_state", {}).get("preconditioning_enabled")

    @property
    def is_preconditioning_weekday_only(self) -> bool:
        """Return if preconditioning is weekday only for scheduled departure."""
        return DAY_SELECTION_MAP.get(
            self._vehicle_data.get("charge_state", {}).get("preconditioning_times")
        )

    @property
    def scheduled_charging_mode(self) -> str:
        """Return 'Off', 'DepartBy', or 'StartAt' for schedule disabled, scheduled departure, and scheduled charging respectively."""
        return self._vehicle_data.get("charge_state", {}).get("scheduled_charging_mode")

    @property
    def is_scheduled_charging_pending(self) -> bool:
        """Return if scheduled charging is pending."""
        return self._vehicle_data.get("charge_state", {}).get(
            "scheduled_charging_pending"
        )

    @property
    def scheduled_charging_start_time_app(self) -> int:
        """Return the scheduled charging start time."""
        return self._vehicle_data.get("charge_state", {}).get(
            "scheduled_charging_start_time_app"
        )

    async def _send_command(
        self,
        name: str,
        *,
        additional_path_vars: dict = None,
        wake_if_asleep: bool = True,
        **kwargs,
    ) -> Optional[dict]:
        """Wrap commands sent to Tesla API.

        Args
            name: Name of command to send, from endpoints.json
            additional_path_vars: Additional URI variables ('vehicle_id' already included)
            wake_if_asleep: (default True) Wake car if it's asleep before sending the command
            **kwargs: Any additional parameters for the api call

        """
        path_vars = {"vehicle_id": self.id}
        if additional_path_vars:
            path_vars.update(additional_path_vars)

        _LOGGER.debug("Sending command: %s", name)
        try:
            data = await self._controller.api(
                name, path_vars=path_vars, wake_if_asleep=wake_if_asleep, **kwargs
            )
            _LOGGER.debug("Response from command %s: %s", name, data)
            return data
        except TeslaException as ex:
            if ex.code == 408 and not wake_if_asleep and not self.is_on:
                # 408 due to being asleep and we didn't try to wake it
                _LOGGER.debug(
                    "Vehicle unavailable for command: %s, car state: %s, wake_if_asleep: %s",
                    name,
                    self.state,
                    wake_if_asleep,
                )
                return None
            raise ex

    def _get_lat_long(self) -> Tuple[Optional[float], Optional[float]]:
        """Get current latitude and longitude."""
        lat = None
        long = None

        if self.native_location_supported:
            long = self.native_longitude
            lat = self.native_latitude
        else:
            long = self.longitude
            lat = self.latitude

        return lat, long

    def _get_active_route_lat_long(self) -> Tuple[Optional[float], Optional[float]]:
        """Get active route latitude and longitude."""
        lat = None
        long = None

        if self.active_route_longitude and self.active_route_latitude:
            lat = self.active_route_latitude
            long = self.active_route_longitude

        return lat, long

    def update_car_info(self, car: dict) -> None:
        """Update the car info dict from the vehicle_list api."""
        if not car:
            _LOGGER.debug("Attempted to update car id %d with empty car info", self.id)
            return
        if car["vin"] != self.vin:
            _LOGGER.error(
                "Failed updating car info: new VIN (%s) doesn't match existing vin (%s)",
                car["vin"][-5:],
                self.vin[-5:],
            )
            return
        self._car.update(car)

    async def change_charge_limit(self, value: float) -> None:
        """Send command to change charge limit."""
        # Only wake car if the value is different
        wake_if_asleep = value != self.charge_limit_soc
        data = await self._send_command(
            "CHANGE_CHARGE_LIMIT", percent=int(value), wake_if_asleep=wake_if_asleep
        )

        if data and data["response"]["result"] is True:
            params = {"charge_limit_soc": int(value)}
            self._vehicle_data["charge_state"].update(params)

    async def charge_port_door_close(self) -> None:
        """Send command to close charge port door."""
        data = await self._send_command("CHARGE_PORT_DOOR_CLOSE")

        if data and data["response"]["result"] is True:
            params = {"charge_port_door_open": False}
            self._vehicle_data["charge_state"].update(params)

    async def charge_port_door_open(self) -> None:
        """Send command to open charge port door."""
        data = await self._send_command("CHARGE_PORT_DOOR_OPEN")

        if data and data["response"]["result"] is True:
            params = {"charge_port_door_open": True}
            self._vehicle_data["charge_state"].update(params)

    async def flash_lights(self) -> None:
        """Send command to flash lights."""
        await self._send_command("FLASH_LIGHTS", on=True)

    async def honk_horn(self) -> None:
        """Send command to honk horn."""
        await self._send_command("HONK_HORN", on=True)

    async def lock(self) -> None:
        """Send lock command."""
        data = await self._send_command("LOCK")
        if data and data["response"]["result"] is True:
            params = {"locked": True}
            self._vehicle_data["vehicle_state"].update(params)

    async def remote_seat_heater_request(self, level: int, seat_id: int) -> None:
        """Send command to change seat heat.

        Args
            level: 0 (off), 1 (low), 2 (medium), 3 (high)
            seat_id: 0 (front left), 1 (front right), 2 (rear left), 4 (rear center)
            5 (rear right), 6 (third row left), 7 (third row right)

        """
        # If car is asleep the heater is already off
        wake_if_asleep = level > 0
        data = await self._send_command(
            "REMOTE_SEAT_HEATER_REQUEST",
            seat_position=seat_id,
            level=level,
            wake_if_asleep=wake_if_asleep,
        )
        if data and data["response"]["result"] is True:
            params = {f"seat_heater_{SEAT_ID_MAP[seat_id]}": level}
            self._vehicle_data["climate_state"].update(params)

    def get_seat_heater_status(self, seat_id: int) -> int:
        """Return status of seat heater for a given seat."""
        seat_id = f"seat_heater_{SEAT_ID_MAP[seat_id]}"
        if self.data_available:
            return self._vehicle_data.get("climate_state", {}).get(seat_id)
        return None
    
    async def remote_seat_cooler_request(self, level: int, seat_id: int) -> None:
        """Send command to change seat cooler.

        Args
            level: 1 (off), 2 (low), 3 (medium), 4 (high)
            seat_id: 1 (front left), 2 (front right)

        """
        # If car is asleep the cooler is already off
        wake_if_asleep = level > 1
        data = await self._send_command(
            "REMOTE_SEAT_COOLING_REQUEST",
            seat_position=seat_id,
            seat_cooler_level=level,
            wake_if_asleep=wake_if_asleep,
        )
        if data and data["response"]["result"] is True:
            params = {f"seat_fan_front_{SEAT_ID_MAP[seat_id]}": level}
            self._vehicle_data["climate_state"].update(params)
    
    def get_seat_cooler_status(self, seat_id: int) -> int:
        """Return status of seat heater for a given seat."""
        seat_id = f"seat_fan_front_{SEAT_ID_MAP[seat_id]}"
        if self.data_available:
            return self._vehicle_data.get("climate_state", {}).get(seat_id)
        return None

    async def schedule_software_update(self, offset_sec: Optional[int] = 0) -> None:
        """Send command to install software update."""
        await self._send_command("SCHEDULE_SOFTWARE_UPDATE", offset_sec=offset_sec)

    async def set_charging_amps(self, value: float) -> None:
        """Send command to set charging amps."""
        # Only wake car if the value is different
        wake_if_asleep = value != self._vehicle_data.get("charge_state", {}).get(
            "charge_current_request"
        )
        data = await self._send_command(
            "CHARGING_AMPS", charging_amps=int(value), wake_if_asleep=wake_if_asleep
        )
        # A second API call allows setting below 5 Amps
        if value < 5:
            data = await self._send_command(
                "CHARGING_AMPS", charging_amps=int(value), wake_if_asleep=wake_if_asleep
            )

        if data and data["response"]["result"] is True:
            params = {
                "charge_amps": int(value),
                "charge_current_request": int(value),
            }
            self._vehicle_data["charge_state"].update(params)

    async def set_cabin_overheat_protection(self, option: str) -> None:
        """Send command to set cabin overheat protection.

        Args
            option: "Off", "No A/C", "On"

        """

        if option == "Off":
            body_on = False
            fan_only = False
        elif option == "No A/C":
            body_on = True
            fan_only = True
        elif option == "On":
            body_on = True
            fan_only = False

        data = await self._send_command(
            "SET_CABIN_OVERHEAT_PROTECTION", on=body_on, fan_only=fan_only
        )
        if data and data["response"]["result"] is True:
            params = {"cabin_overheat_protection": option}
            self._vehicle_data["climate_state"].update(params)

    async def set_climate_keeper_mode(self, keeper_id: int) -> None:
        """Send command to set climate keeper mode.

        Args
            keeper_id: 1 (keep on), 2 (dog mode), 3 (camp mode)

        """
        # If car is asleep, climate is already off
        wake_if_asleep = keeper_id > 0
        data = await self._send_command(
            "SET_CLIMATE_KEEPER_MODE",
            climate_keeper_mode=keeper_id,
            wake_if_asleep=wake_if_asleep,
        )
        if data and data["response"]["result"] is True:
            params = {
                "climate_keeper_mode": CLIMATE_KEEPER_ID_MAP[keeper_id],
                "is_climate_on": True,
            }
            self._vehicle_data["climate_state"].update(params)

    async def set_bioweapon_mode(self, enable: bool) -> None:
        """Send command to set bioweapon mode.

        Args
            enable: 'True' to enable, 'False' to disable

        """
        # If car is asleep, bioweapon is already off
        data = await self._send_command(
            "HVAC_BIOWEAPON_MODE",
            on=enable,
            manual_override=True,
        )
        if data and data["response"]["result"] is True:
            params = {
                "bioweapon_mode": enable,
                "is_climate_on": True,
            }
            self._vehicle_data["climate_state"].update(params)
    
    async def remote_auto_seat_climate_request(
        self, seat_id: int, enable: bool
    ) -> None:
        """Send command to change seat climate to auto.

        Args
            seat_id: 0 (front left), 1 (front right)
            enable: 'True' to enable, 'False' to disable

        """

        data = await self._send_command(
            "REMOTE_AUTO_SEAT_CLIMATE_REQUEST",
            auto_seat_position=seat_id,
            seat_position=seat_id,
            auto_climate_on=enable,
        )
        if data and data["response"]["result"] is True:
            params = {f"auto_seat_climate_{SEAT_ID_MAP[seat_id]}": enable}
            self._vehicle_data["climate_state"].update(params)

    async def remote_auto_steering_wheel_heat_climate_request(
        self, enable: bool
    ) -> None:
        """Send command to enable or disable auto steering wheel heat.

        Args
            enable: True to enable auto steering wheel heat, False to disable.

        """

        data = await self._send_command(
            "REMOTE_AUTO_STEERING_WHEEL_HEAT_CLIMATE_REQUEST", on=enable
        )
        if data and data["response"]["result"] is True:
            params = {"auto_steering_wheel_heat": enable}
            self._vehicle_data["climate_state"].update(params)

    async def set_heated_steering_wheel(self, value: bool) -> None:
        """Send command to set heated steering wheel."""
        data = await self._send_command(
            "REMOTE_STEERING_WHEEL_HEATER_REQUEST", on=value
        )

        if data and data["response"]["result"] is True:
            params = {"steering_wheel_heater": value}
            self._vehicle_data["climate_state"].update(params)

    async def set_heated_steering_wheel_level(self, level: int) -> None:
        """Send command to set the heated steering wheel level."""
        data = await self._send_command(
            "REMOTE_STEERING_WHEEL_HEAT_LEVEL_REQUEST", level=level
        )

        if data and data["response"]["result"] is True:
            params = {"steering_wheel_heat_level": level}
            self._vehicle_data["climate_state"].update(params)

    def get_heated_steering_wheel_level(self) -> int:
        """Return the status of the heated steering wheel."""
        if self.data_available:
            return self._vehicle_data.get("climate_state", {}).get(
                "steering_wheel_heat_level"
            )
        return None

    async def set_hvac_mode(self, value: str) -> None:
        """Send command to set HVAC mode.

        Args
            value: on, off

        """
        if value == "off":
            # If car is asleep, climate is already off
            data = await self._send_command("CLIMATE_OFF", wake_if_asleep=False)
            if data and data["response"]["result"] is True:
                # Set additional values if turning HVAC off after defrost max
                params = {
                    "defrost_mode": 0,
                    "driver_temp_setting": self._previous_driver_temp,
                    "is_climate_on": False,
                    "is_front_defroster_on": False,
                    "is_rear_defroster_on": False,
                    "passenger_temp_setting": self._previous_passenger_temp,
                }
                self._vehicle_data["climate_state"].update(params)

        elif value == "on":
            data = await self._send_command("CLIMATE_ON")
            if data and data["response"]["result"] is True:
                params = {"is_climate_on": True}
                self._vehicle_data["climate_state"].update(params)

    async def set_max_defrost(self, state: int) -> None:
        """Send command to set max defrost.

        Args
            state: 2 = on, 0 = off

        """
        # If car is asleep, climate is already off
        wake_if_asleep = state > 0
        data = await self._send_command(
            "MAX_DEFROST", on=state, wake_if_asleep=wake_if_asleep
        )
        if data and data["response"]["result"] is True:
            self._previous_driver_temp = self.driver_temp_setting
            self._previous_passenger_temp = self.passenger_temp_setting
            self._previous_fan_status = self.fan_status

            if state == 2:
                params = {
                    "defrost_mode": state,
                    "driver_temp_setting": self.max_avail_temp,
                    "fan_status": 7,
                    "is_climate_on": True,
                    "is_front_defroster_on": True,
                    "is_rear_defroster_on": True,
                    "passenger_temp_setting": self.max_avail_temp,
                }
            if state == 0:
                params = {
                    "defrost_mode": state,
                    "driver_temp_setting": self._previous_driver_temp,
                    "fan_status": self._previous_fan_status,
                    "is_climate_on": True,
                    "is_front_defroster_on": False,
                    "is_rear_defroster_on": False,
                    "passenger_temp_setting": self._previous_passenger_temp,
                }
            self._vehicle_data["climate_state"].update(params)

    async def set_sentry_mode(self, value: bool) -> None:
        """Send command to set sentry mode."""
        # If car is asleep, sentry is already off
        data = await self._send_command(
            "SET_SENTRY_MODE", on=value, wake_if_asleep=value
        )

        if data and data["response"]["result"] is True:
            params = {"sentry_mode": value}
            self._vehicle_data["vehicle_state"].update(params)

    async def set_temperature(self, temp: float) -> None:
        """Send command to set temperature."""
        data = await self._send_command(
            "CHANGE_CLIMATE_TEMPERATURE_SETTING", driver_temp=temp, passenger_temp=temp
        )
        if data and data["response"]["result"] is True:
            params = {"driver_temp_setting": temp}
            self._vehicle_data["climate_state"].update(params)

    async def start_charge(self) -> None:
        """Send command to start charge."""
        data = await self._send_command("START_CHARGE")

        if data and data["response"]["result"] is True:
            params = {"charging_state": "Charging"}
            self._vehicle_data["charge_state"].update(params)

    async def stop_charge(self) -> None:
        """Send command to stop charge."""
        # If car is asleep, it's not charging
        data = await self._send_command("STOP_CHARGE", wake_if_asleep=False)

        if data and data["response"]["result"] is True:
            params = {"charging_state": "Stopped"}
            self._vehicle_data["charge_state"].update(params)

    async def wake_up(self) -> None:
        """Send command to wake up."""
        await self._controller.wake_up(car_id=self.id)

    async def toggle_trunk(self) -> None:
        """Actuate rear trunk."""
        prev_is_trunk_closed = self.is_trunk_closed
        data = await self._send_command("ACTUATE_TRUNK", which_trunk="rear")
        if data and data["response"]["result"] is True:
            if not prev_is_trunk_closed:
                params = {"rt": 0}
                self._vehicle_data["vehicle_state"].update(params)
            if prev_is_trunk_closed:
                params = {"rt": 255}
                self._vehicle_data["vehicle_state"].update(params)

    async def toggle_frunk(self) -> None:
        """Actuate front trunk."""
        prev_is_frunk_closed = self.is_frunk_closed
        data = await self._send_command("ACTUATE_TRUNK", which_trunk="front")
        if data and data["response"]["result"] is True:
            if not prev_is_frunk_closed:
                params = {"ft": 0}
                self._vehicle_data["vehicle_state"].update(params)
            if prev_is_frunk_closed:
                params = {"ft": 255}
                self._vehicle_data["vehicle_state"].update(params)

    async def trigger_homelink(self) -> None:
        """Send command to trigger homelink."""
        if self.homelink_device_count is None:
            raise HomelinkError(f"No homelink devices added to {self.display_name}.")

        if self.homelink_nearby is not True:
            raise HomelinkError(f"No homelink devices near {self.display_name}.")

        lat, long = self._get_lat_long()

        data = await self._send_command("TRIGGER_HOMELINK", lat=lat, lon=long)

        if data and data["response"]:
            _LOGGER.debug("Homelink response: %s", data["response"])
            result = data["response"]["result"]
            reason = data["response"]["reason"]
            if result is False:
                raise HomelinkError(f"Error calling trigger_homelink: {reason}")

    async def unlock(self) -> None:
        """Send unlock command."""
        data = await self._send_command("UNLOCK")
        if data and data["response"]["result"] is True:
            params = {"locked": False}
            self._vehicle_data["vehicle_state"].update(params)

    async def vent_windows(self) -> None:
        """Vent Windows."""
        data = await self._send_command("WINDOW_CONTROL", command="vent", lat=0, long=0)
        if data and data["response"]["result"] is True:
            params = {
                "fd_window": 1,
                "fp_window": 1,
                "rd_window": 1,
                "rp_window": 1,
            }
            self._vehicle_data["vehicle_state"].update(params)

    async def close_windows(self) -> None:
        """Close Windows."""
        data = await self._send_command(
            "WINDOW_CONTROL", command="close", lat=self.latitude, long=self.longitude
        )
        if data and data["response"]["result"] is True:
            params = {
                "fd_window": 0,
                "fp_window": 0,
                "rd_window": 0,
                "rp_window": 0,
            }
            self._vehicle_data["vehicle_state"].update(params)

    async def valet_mode(self, enable, pin=None) -> None:
        """Set Valet Mode.

        Args
            enable: True to activate, False to deactivate.
            pin: optional, pin not required to activate or deactivate valet mode.
                Even with a previous PIN set. If you clear the PIN and activate Valet Mode
                without the parameter, you will only be able to deactivate it from your
                car's screen by signing into your Tesla account.

        """
        if pin:
            data = await self._send_command("SET_VALET_MODE", on=enable, password=pin)
        else:
            data = await self._send_command("SET_VALET_MODE", on=enable)

        if data and data["response"]:
            _LOGGER.debug("Valet mode response: %s", data["response"])
            result = data["response"]["result"]
            reason = data["response"]["reason"]
            if result is False:
                _LOGGER.debug("Error calling valet mode: %s", reason)
            else:
                if enable:
                    params = {"valet_mode": True}
                else:
                    params = {"valet_mode": False}
                self._vehicle_data["vehicle_state"].update(params)

    async def remote_start(self) -> None:
        """Remote start."""
        data = await self._send_command("REMOTE_START")

        if data and data["response"]:
            _LOGGER.debug("Remote start response: %s", data["response"])
            result = data["response"]["result"]
            reason = data["response"]["reason"]
            if result is False:
                _LOGGER.debug("Error calling remote start: %s", reason)
            else:
                self._vehicle_data["vehicle_state"].update({"remote_start": True})

    async def set_scheduled_departure(
        self,
        enable: bool,
        departure_time: int,
        preconditioning_enabled: bool,
        preconditioning_weekdays_only: bool,
        off_peak_charging_enabled: bool,
        off_peak_charging_weekdays_only: bool,
        end_off_peak_time: int,
    ) -> None:
        """Send command to set departure time.

        Args
            enable: Turn on (True) or turn off (False) the scheduled departure.
            departure_time: Time in minutes after midnight (local time) for the departure.
            preconditioning_enabled: Enable (True) or disable (False) the climate preconditioning.
            preconditioning_weekdays_only: Precondition climate for departure time on weekdays only (True) or all days (False).
            off_peak_charging_enabled: Complete charging during off peak hours (True) or complete charging just before departure time (False).
            off_peak_charging_weekdays_only: Complete off peak charging only on weekdays only (True) or all days (False).
            end_off_peak_time: Time in minutes after midnight when the off peak rate ends.

        """
        data = await self._send_command(
            "SCHEDULED_DEPARTURE",
            enable=enable,
            departure_time=departure_time,
            preconditioning_enabled=preconditioning_enabled,
            preconditioning_weekdays_only=preconditioning_weekdays_only,
            off_peak_charging_enabled=off_peak_charging_enabled,
            off_peak_charging_weekdays_only=off_peak_charging_weekdays_only,
            end_off_peak_time=end_off_peak_time,
        )

        if data and data["response"]["result"] is True:
            if enable:
                mode_str = "DepartBy"
            else:
                mode_str = "Off"

            params = {
                "scheduled_charging_mode": mode_str,
                "scheduled_departure_time_minutes": departure_time,
                "preconditioning_enabled": preconditioning_enabled,
                "preconditioning_weekdays_only": list(DAY_SELECTION_MAP.values()).index(
                    preconditioning_weekdays_only
                ),
                "off_peak_charging_enabled": off_peak_charging_enabled,
                "off_peak_charging_weekdays_only": list(
                    DAY_SELECTION_MAP.values()
                ).index(off_peak_charging_weekdays_only),
                "end_off_peak_time": end_off_peak_time,
            }
            self._vehicle_data["charge_state"].update(params)

    async def set_scheduled_charging(self, enable: bool, time: int) -> None:
        """Send command to set scheduled charging time.

        Args
            enable: Turn on (True) or turn off (False) the scheduled charging.
            time: Time in minutes after midnight (local time) to start charging.

        """
        data = await self._send_command("SCHEDULED_CHARGING", enable=enable, time=time)

        if data and data["response"]["result"] is True:
            if enable:
                mode_str = "StartAt"
            else:
                mode_str = "Off"
                time = None
            params = {
                "scheduled_charging_mode": mode_str,
                "scheduled_charging_start_time": time,
                "scheduled_charging_pending": enable,
            }
            self._vehicle_data["charge_state"].update(params)

    async def remote_boombox(self) -> None:
        """Remote boombox."""
        data = await self._send_command("REMOTE_BOOMBOX")

        if data and data["response"]:
            result = data["response"]["result"]
            reason = data["response"]["reason"]
            if result is False:
                _LOGGER.debug("Error calling remote boombox: %s", reason)
            else:
                _LOGGER.debug("Remote boombox called successfully.")



#  SPDX-License-Identifier: Apache-2.0
"""
Python Package for controlling Tesla API.

For more details about this api, please refer to the documentation at
https://github.com/zabuldon/teslajsonpy
"""
import asyncio
import logging
import pkgutil
import time
import ssl
from json import JSONDecodeError
from typing import Dict, List, Optional, Set, Text

import httpx
import orjson
from tenacity import retry, stop_after_delay
from yarl import URL

from teslajsonpy.car import TeslaCar
from teslajsonpy.connection import Connection
from teslajsonpy.const import (
    AUTH_DOMAIN,
    DRIVING_INTERVAL,
    IDLE_INTERVAL,
    MAX_API_RETRY_TIME,
    ONLINE_INTERVAL,
    RESOURCE_TYPE,
    RESOURCE_TYPE_BATTERY,
    RESOURCE_TYPE_SOLAR,
    SLEEP_INTERVAL,
    UPDATE_INTERVAL,
    WAKE_CHECK_INTERVAL,
    WAKE_TIMEOUT,
    CLIENT_ID,
)
from teslajsonpy.energy import EnergySite, PowerwallSite, SolarPowerwallSite, SolarSite
from teslajsonpy.exceptions import (
    TeslaException,
    custom_retry,
    custom_retry_except_unavailable,
    custom_wait,
)

_LOGGER = logging.getLogger(__name__)


def valid_result(result):
    """Check if TeslaAPI result successful.

    Parameters
    ----------
    result : tesla API result
        This is the result of a Tesla Rest API call.

    Returns
    -------
    bool
        Tesla API failure can be checked in a dict with a bool in
        ['response']['result'], a bool, or None or
        ['response']['reason'] == 'could_not_wake_buses'
        Returns true when a failure state not detected.

    """
    try:
        return (
            result is not None
            and result is not False
            and (
                result is True
                or (
                    isinstance(result, dict)
                    and (
                        (
                            isinstance(result["response"], dict)
                            and (
                                result["response"].get("result") is True
                                or result["response"].get("reason")
                                != "could_not_wake_buses"
                            )
                        )
                        or isinstance(result.get("response"), list)
                    )
                )
            )
        )
    except TypeError as exception:
        _LOGGER.error("Result: %s, %s", result, exception)
        return False


class Controller:
    #  pylint: disable=too-many-public-methods
    """Controller for connections to Tesla Motors API."""

    def __init__(
        self,
        websession: Optional[httpx.AsyncClient] = None,
        email: Text = None,
        password: Text = None,
        access_token: Text = None,
        refresh_token: Text = None,
        expiration: int = 0,
        update_interval: int = UPDATE_INTERVAL,
        enable_websocket: bool = False,
        polling_policy: Text = None,
        auth_domain: str = AUTH_DOMAIN,
        api_proxy_url: str = None,
        api_proxy_cert: str = None,
        client_id: str = CLIENT_ID
    ) -> None:
        """Initialize controller.

        Args:
            websession (aiohttp.ClientSession): Websession for aiohttp.
            email (Text, optional): Email account. Defaults to None.
            password (Text, optional): Password. Defaults to None.
            access_token (Text, optional): Access token. Defaults to None.
            refresh_token (Text, optional): Refresh token. Defaults to None.
            expiration (int, optional): Timestamp when access_token expires. Defaults to 0
            update_interval (int, optional): Seconds between allowed updates to the API.  This is to prevent
            being blocked by Tesla. Defaults to UPDATE_INTERVAL.
            enable_websocket (bool, optional): Whether to connect with websockets. Defaults to False.
            polling_policy (Text, optional): How aggressively will we poll the car. Possible values:
            Not set - Only keep the car awake while it is actively charging or driving, and while sentry
            mode is enabled (default).
            'connected' - Also keep the car awake while it is connected to a charger, even if the charging
            session is complete.
            'always' - Keep polling the car at all times.  Will possibly never allow the car to sleep.
            auth_domain (str, optional): The authentication domain. Defaults to const.AUTH_DOMAIN
            api_proxy_url (str, optional): HTTPS Proxy for Fleet API commands
            api_proxy_cert (str, optional): Custom SSL certificate to use with proxy
            client_id (str, optional): Required for modern vehicles using Fleet API

        """
        if not websession or not isinstance(websession, httpx.AsyncClient):
            # create_default_context() does blocking I/O. It is recommended
            # to always pass an httpx.AsyncClient instance to the Controller
            ssl_context = ssl.create_default_context()
            websession = httpx.AsyncClient(timeout=60, verify=ssl_context)

        if api_proxy_cert:
            # Loading custom SSL certificate for proxy does blocking I/O.
            # It is recommended to instead pass an httpx.AsyncClient that
            # already has an SSL context with the custom certificate loaded.
            try:
                ssl_context.load_verify_locations(api_proxy_cert)
            except (FileNotFoundError, ssl.SSLError):
                _LOGGER.warning("Unable to load custom SSL certificate from %s", api_proxy_cert)

        self.__connection = Connection(
            websession=websession,
            email=email,
            password=password,
            access_token=access_token,
            refresh_token=refresh_token,
            expiration=expiration,
            auth_domain=auth_domain,
            client_id=client_id,
            api_proxy_url=api_proxy_url,
        )
        self._update_interval: int = update_interval
        self._update_interval_vin = {}
        self.__update = {}
        self.__driving = {}  # for websocket timestamp only
        self._last_update_time = {}  # succesful update attempts by car
        self._last_wake_up_time = {}  # succesful wake_ups by car
        self._last_attempted_update_time = 0  # all attempts by controller
        self.__lock = {}
        self.__update_lock = None  # controls access to update function
        self.__wakeup_lock = {}
        self.car_online = {}
        self.__id_vin_map = {}
        self.__vin_id_map = {}
        self.__vin_vehicle_id_map = {}
        self.__vehicle_id_vin_map = {}
        self.__websocket_listeners = []
        self.__last_parked_timestamp = {}
        self.__update_state = {}
        self.enable_websocket = enable_websocket
        self.endpoints = {}
        self.polling_policy = polling_policy

        self._include_vehicles: bool = True
        self._include_energysites: bool = True
        self._product_list: List[dict] = []
        self._vehicle_list: List[dict] = []
        self._vehicle_data: Dict[str, dict] = {}
        self._energysite_list: List[dict] = []
        self._site_config: Dict[int, dict] = {}
        self._site_data: Dict[int, dict] = {}
        self._site_summary: Dict[int, dict] = {}
        self._grid_status_unknown: Dict[int, bool] = {}
        self.cars: Dict[str, TeslaCar] = {}
        self.energysites: Dict[int, EnergySite] = {}

    async def connect(
        self,
        test_login: bool = False,
        include_vehicles: bool = True,
        include_energysites: bool = True,
        mfa_code: Text = "",
    ) -> Dict[Text, Text]:
        """Connect controller to Tesla.

        Args
            test_login (bool, optional): Whether to test credentials only. Defaults to False.
            include_vehicles (bool, optional): Whether to include vehicles. Defaults to True.
            include_energysites(bool, optional): Whether to include energysites. Defaults to True.
            mfa_code (Text, optional): MFA code to use for connection

        Returns
            Dict[Text, Text]: Returns the refresh_token, access_token, id_token and expires_in time

        """

        if mfa_code:
            self.__connection.mfa_code = mfa_code

        self._last_attempted_update_time = round(time.time())
        self.__update_lock = asyncio.Lock()

        if not test_login:
            self._product_list = await self.get_product_list()

            if include_vehicles:
                self._vehicle_list = [
                    cars for cars in self._product_list if "vehicle_id" in cars
                ]

            if include_energysites:
                self._energysite_list = [
                    p
                    for p in self._product_list
                    if p.get(RESOURCE_TYPE) == RESOURCE_TYPE_SOLAR
                    or p.get(RESOURCE_TYPE) == RESOURCE_TYPE_BATTERY
                ]

        return {
            "refresh_token": self.__connection.refresh_token,
            "access_token": self.__connection.access_token,
            "expiration": self.__connection.expiration,
            "id_token": self.__connection.id_token,
        }

    async def disconnect(self) -> None:
        """Disconnect from Tesla api."""
        _LOGGER.debug("Disconnecting controller.")
        await self.__connection.close()

    def is_token_refreshed(self) -> bool:
        """Return whether token has been changed and not retrieved.

        Returns
            bool: Whether token has been changed since the last return

        """
        return self.__connection.token_refreshed

    def get_tokens(self) -> Dict[Text, Text]:
        """Return oauth data including refresh and access tokens, and expires time.

        This will set the the self.__connection token_refreshed to False.

        Returns
            Dict[Text, Text]: Returns the refresh_token, access_token, id_token and expires time

        """
        self.__connection.token_refreshed = False
        return {
            "refresh_token": self.__connection.refresh_token,
            "access_token": self.__connection.access_token,
            "expiration": self.__connection.expiration,
            "id_token": self.__connection.id_token,
        }

    def get_expiration(self) -> int:
        """Return expiration for oauth.

        Returns
            int: Returns timestamp when oauth expires

        """
        return self.__connection.expiration

    def get_oauth_url(self) -> URL:
        """Return oauth url."""
        return self.__connection.get_authorization_code_link(new=True)

    def set_authorization_code(self, code: Text) -> None:
        """Set authorization code in Connection."""
        self.__connection.code = code

    def set_authorization_domain(self, domain: Text) -> None:
        """Set authorization domain in Connection."""
        if not domain:
            return
        if self.__connection.auth_domain.host != domain:
            self.__connection.auth_domain = self.__connection.auth_domain.with_host(
                domain
            )

    def register_websocket_callback(self, callback) -> int:
        """Register callback for websocket messages.

        Args
            callback (function): function to call with json data

        Returns
            int: Return index of entry

        """
        self.__websocket_listeners.append(callback)
        return len(self.__websocket_listeners) - 1

    async def get_product_list(self) -> list:
        """Get product list from Tesla."""
        return (await self.api("PRODUCT_LIST"))["response"]

    async def get_site_config(self, energysite_id: int) -> dict:
        """Get site config json from TeslaAPI for a given energysite_id."""
        return (await self.api("SITE_CONFIG", path_vars={"site_id": energysite_id}))[
            "response"
        ]

    async def get_vehicle_data(self, vin: str, wake_if_asleep: bool = False) -> dict:
        """Get vehicle data json from TeslaAPI for a given vin."""
        try:
            response = (
                await self.api(
                    "VEHICLE_DATA",
                    path_vars={"vehicle_id": self._vin_to_id(vin)},
                    wake_if_asleep=wake_if_asleep,
                )
            )["response"]

        except TeslaException as ex:
            if ex.message == "VEHICLE_UNAVAILABLE":
                _LOGGER.debug("Vehicle offline - data unavailable.")
                return {}
            raise ex

        return response

    async def get_vehicle_summary(self, vin: str) -> dict:
        """Get vehicle summary json from TeslaAPI for a given vin."""
        return (
            await self.api(
                "VEHICLE_SUMMARY",
                path_vars={"vehicle_id": self._vin_to_id(vin)},
                wake_if_asleep=False,
            )
        )["response"]

    async def get_site_data(self, energysite_id: int) -> dict:
        """Get site data json from TeslaAPI for a given energysite_id."""
        return (await self.api("SITE_DATA", path_vars={"site_id": energysite_id}))[
            "response"
        ]

    async def get_site_summary(self, energysite_id: int) -> dict:
        """Get site data json from TeslaAPI for a given energysite_id."""
        return (await self.api("SITE_SUMMARY", path_vars={"site_id": energysite_id}))[
            "response"
        ]

    async def generate_car_objects(
        self,
        wake_if_asleep: bool = False,
        filtered_vins: Optional[List[Text]] = None,
    ) -> Dict[str, TeslaCar]:
        """Generate car objects.

        Args
            wake_if_asleep (bool, optional): Wake up vehicles if asleep.
            filtered_vins (list, optional): If not empty, filters the cars by the provided VINs.

        """
        for car in self._vehicle_list:
            vin = car["vin"]
            if filtered_vins and vin not in filtered_vins:
                _LOGGER.debug("Skipping car with VIN: %s", vin)
                continue

            self.set_id_vin(car_id=car["id"], vin=vin)
            self.set_vehicle_id_vin(vehicle_id=car["vehicle_id"], vin=vin)
            self.__lock[vin] = asyncio.Lock()
            self.__wakeup_lock[vin] = asyncio.Lock()
            self._last_update_time[vin] = 0
            self._last_wake_up_time[vin] = 0
            self.__update[vin] = True
            self.__update_state[vin] = "normal"
            self.set_car_online(vin=vin, online_status=car["state"] == "online")
            self.set_last_park_time(vin=vin, timestamp=self._last_attempted_update_time)
            self.__driving[vin] = {}
            self._vehicle_data[vin] = {}

            try:
                self._vehicle_data[vin] = await self.get_vehicle_data(
                    vin, wake_if_asleep=wake_if_asleep
                )
            except TeslaException as ex:
                _LOGGER.warning(
                    "Unable to get vehicle data during setup, car will still be added. %s: %s",
                    ex.code,
                    ex.message,
                )
            self.cars[vin] = TeslaCar(car, self, self._vehicle_data[vin])

        return self.cars

    async def generate_energysite_objects(self) -> Dict[int, EnergySite]:
        """Generate energy site objects."""
        for energysite in self._energysite_list:
            energysite_id = energysite["energy_site_id"]

            self._site_config[energysite_id] = await self.get_site_config(energysite_id)
            # For dealing with sites that always report "Unknown"
            # Default to True and check during updates
            self._grid_status_unknown = {energysite_id: True}
            # Solar only systems (no Powerwalls) are listed as "solar"
            try:
                self._site_data[energysite_id] = await self.get_site_data(
                    energysite_id
                )
            except TeslaException as ex:
                _LOGGER.warning(
                    "Unable to get site data during setup, site will still be added. %s: %s",
                    ex.code,
                    ex.message,
                )
                self._site_data[energysite_id] = {}

            if energysite[RESOURCE_TYPE] == RESOURCE_TYPE_SOLAR:

                self.energysites[energysite_id] = SolarSite(
                    self.api,
                    energysite,
                    self._site_config[energysite_id],
                    self._site_data[energysite_id],
                )
            # Powerwall systems listed as "battery"
            if energysite[RESOURCE_TYPE] == RESOURCE_TYPE_BATTERY:

                self._site_summary[energysite_id] = await self.get_site_summary(
                    energysite_id
                )

                if energysite["components"]["solar"]:
                    self.energysites[energysite_id] = SolarPowerwallSite(
                        self.api,
                        energysite,
                        self._site_config[energysite_id],
                        self._site_data[energysite_id],
                        self._site_summary[energysite_id],
                    )
                else:
                    self.energysites[energysite_id] = PowerwallSite(
                        self.api,
                        energysite,
                        self._site_config[energysite_id],
                        self._site_data[energysite_id],
                        self._site_summary[energysite_id],
                    )

        return self.energysites

    async def wake_up(self, car_id) -> bool:
        """Attempt to wake the car, returns True if successfully awakened."""
        car_vin = self._id_to_vin(car_id)
        car_id = self._update_id(car_id)
        async with self.__wakeup_lock[car_vin]:
            wake_start_time = time.time()
            wake_deadline = wake_start_time + WAKE_TIMEOUT
            _LOGGER.debug(
                "%s: Sending wake request with deadline of: %d",
                car_vin[-5:],
                wake_deadline,
            )
            result = await self.api(
                "WAKE_UP", path_vars={"vehicle_id": car_id}, wake_if_asleep=False
            )
            state = result.get("response", {}).get("state")
            self.set_car_online(
                vin=car_vin,
                car_id=car_id,
                online_status=state == "online",
            )
            while not self.is_car_online(vin=car_vin) and time.time() < wake_deadline:
                await asyncio.sleep(WAKE_CHECK_INTERVAL)
                response = await self.get_vehicle_summary(vin=car_vin)
                state = response.get("state")
                self.set_car_online(
                    vin=car_vin,
                    car_id=car_id,
                    online_status=state == "online",
                )

            _LOGGER.debug(
                "%s: Wakeup took %d seconds, state: %s",
                car_vin[-5:],
                time.time() - wake_start_time,
                state,
            )
            return self.is_car_online(
                vin=car_vin,
                car_id=car_id,
            )

    def _calculate_next_interval(self, vin: Text) -> int:
        cur_time = round(time.time())
        _LOGGER.debug(
            "%s: %s. Polling policy: %s. Update state: %s. Since last park: %s. Since last wake up: %s. Idle interval: %s. shift_state: %s sentry: %s climate: %s, charging: %s ",
            vin[-5:],
            self.cars[vin].state,
            self.polling_policy,
            self.__update_state.get(vin),
            cur_time - self.get_last_park_time(vin=vin),
            cur_time - self.get_last_wake_up_time(vin=vin),
            IDLE_INTERVAL,
            self.cars[vin].shift_state,
            self.cars[vin].sentry_mode,
            self.cars[vin].is_climate_on,
            self.cars[vin].charging_state,
        )
        if vin not in self.__update_state:
            self.__update_state[vin] = "normal"

        if self.cars[vin].state == "asleep" or self.cars[vin].shift_state:
            self.set_last_park_time(
                vin=vin, timestamp=cur_time, shift_state=self.cars[vin].shift_state
            )

        if self.cars[vin].is_in_gear:
            driving_interval = min(
                DRIVING_INTERVAL, self.get_update_interval_vin(vin=vin)
            )
            if self.__update_state[vin] != "driving":
                self.__update_state[vin] = "driving"
                _LOGGER.debug(
                    "%s driving; increasing scan rate to every %s seconds",
                    vin[-5:],
                    driving_interval,
                )
            return driving_interval

        if self.polling_policy == "always":
            _LOGGER.debug(
                "%s: %s; Polling policy set to '%s'. Scanning every %s seconds",
                vin[-5:],
                self.cars[vin].state,
                self.polling_policy,
                self.get_update_interval_vin(vin=vin),
            )
            self.__update_state[vin] = "normal"
            return self.get_update_interval_vin(vin=vin)

        if self.polling_policy == "connected" and (
            self.cars[vin].sentry_mode
            or self.cars[vin].is_climate_on
            or (
                self.cars[vin].charging_state
                and self.cars[vin].charging_state != "Disconnected"
                and self.cars[vin].charging_state != ""
            )
            or cur_time - self.get_last_wake_up_time(vin=vin) <= IDLE_INTERVAL
        ):
            _LOGGER.debug(
                "%s: %s; Polling policy set to '%s' and car is connected. "
                "or last_wake_up_time < IDLE_INTERVAL "
                "Polling every %s seconds",
                vin[-5:],
                self.cars[vin].state,
                self.polling_policy,
                self.get_update_interval_vin(vin=vin),
            )
            self.__update_state[vin] = "normal"
            return self.get_update_interval_vin(vin=vin)

        if (cur_time - self.get_last_park_time(vin=vin) > IDLE_INTERVAL) and not (
            self.cars[vin].sentry_mode
            or self.cars[vin].is_climate_on
            or self.cars[vin].charging_state == "Charging"
        ):
            sleep_interval = max(SLEEP_INTERVAL, self.get_update_interval_vin(vin=vin))
            if self.__update_state[vin] != "trying_to_sleep":
                self.__update_state[vin] = "trying_to_sleep"
            _LOGGER.debug(
                "%s: %s; Polling policy set to '%s', trying to sleep; scan throttled to %s seconds and will ignore updates for %s seconds",
                vin[-5:],
                self.cars[vin].state,
                self.polling_policy,
                sleep_interval,
                sleep_interval + self._last_update_time[vin] - cur_time,
            )
            return sleep_interval

        if self.__update_state[vin] != "normal":
            self.__update_state[vin] = "normal"
            _LOGGER.debug(
                "%s: %s; Polling policy set to '%s', scanning every %s seconds",
                vin[-5:],
                self.cars[vin].state,
                self.polling_policy,
                self.get_update_interval_vin(vin=vin),
            )

        return self.get_update_interval_vin(vin=vin)

    async def update(
        self,
        car_id: Optional[Text] = None,
        wake_if_asleep: bool = False,
        force: bool = False,
        update_vehicles: bool = True,
        vins: Optional[Set[str]] = None,
        energy_site_ids: Optional[Set[str]] = None,
    ) -> bool:
        #  pylint: disable=too-many-locals,too-many-statements
        """Update all vehicle and energy site attributes in the cache.

        This command will connect to the Tesla API and first update the list of
        online vehicles assuming no attempt for at least the [update_interval].
        It will then update all the cached values for cars that are awake
        assuming no update has occurred for at least the [update_interval].

        For energy sites, they will only be updated if car_id is blank.

        Args
            car_id (Text, optional): The vehicle to update. If None, all cars are updated. Defaults to None.
            wake_if_asleep (bool, optional): force a vehicle awake. Defaults to False.
            force (bool, optional): force a vehicle update regardless of the update_interval. Defaults to False.
            vins: (Set[str], optional): The cars to update. If None, all cars are updated. Defaults to None.
            energy_site_ids: (Set[str], optional): The energy sites to update. If None, all sites are updated. Defaults to None.

        Returns
            Whether update was successful.

        """
        tasks = []

        async def _get_and_process_car_data(vin: str) -> None:
            async with self.__lock[vin]:
                _LOGGER.debug("%s: Updating VEHICLE_DATA", vin[-5:])
                try:
                    response = await self.get_vehicle_data(
                        vin, wake_if_asleep=wake_if_asleep
                    )
                except TeslaException as ex:
                    # VEHICLE_UNAVAILABLE is handled in get_vehicle_data as debug and ignore
                    # Anything else would be caught here and logged as a warning
                    _LOGGER.warning(
                        "Unable to get vehicle data during poll. %s: %s",
                        ex.code,
                        ex.message,
                    )
                    response = None

                if response:
                    if (
                        self.cars[vin].is_climate_on
                        and self.cars[vin].is_climate_on
                        != response["drive_state"]["shift_state"]
                        and (
                            response["drive_state"]["shift_state"] is None
                            or response["drive_state"]["shift_state"] == "P"
                        )
                    ):
                        self.set_last_park_time(
                            vin=vin,
                            timestamp=response["drive_state"]["timestamp"] / 1000,
                            shift_state=response["drive_state"]["shift_state"],
                        )
                    self._last_update_time[vin] = round(time.time())

                    if self.enable_websocket and self.cars[vin].is_in_gear:
                        asyncio.create_task(
                            self.__connection.websocket_connect(
                                vin[-5:],
                                self.vin_to_vehicle_id(vin=vin),
                                on_message=self._process_websocket_message,
                                on_disconnect=self._process_websocket_disconnect,
                            )
                        )

                    self._vehicle_data[vin].update(response)

        async def _get_and_process_site_data(energysite_id: int) -> None:
            _LOGGER.debug("Updating SITE_DATA for energysite: %s", energysite_id)
            try:
                response = await self.get_site_data(energysite_id)
            except TeslaException:
                response = None

            if response:
                # Some setups always report grid_status of "Unknown" regardless
                # of the actual grid status. Others only report grid_status "Unknown"
                # when the actual grid status is unknown. These setups also sometimes
                # report an incorrect solar_power value of 0.
                if (
                    "grid_status" not in response
                    or response.get("grid_status") != "Unknown"
                ):
                    self._grid_status_unknown[energysite_id] = False

                if (
                    energysite_id in self._grid_status_unknown
                    and not self._grid_status_unknown[energysite_id]
                    and (
                        response.get("grid_status") == "Unknown"
                        and response.get("solar_power") == 0
                    )
                ):
                    _LOGGER.debug(
                        "Ignoring possible spurious energy site solar power read."
                    )
                    del response["solar_power"]

                self._site_data[energysite_id].update(response)

        async def _get_and_process_site_summary(energysite_id: int) -> None:
            _LOGGER.debug("Updating SITE_SUMMARY for energysite: %s", energysite_id)
            try:
                response = await self.get_site_summary(energysite_id)
            except TeslaException:
                response = None

            if response:
                self._site_summary[energysite_id].update(response)

        async def _get_and_process_site_config(energysite_id: int) -> None:
            _LOGGER.debug("Updating SITE_CONFIG for energysite: %s", energysite_id)
            try:
                response = await self.get_site_config(energysite_id)
            except TeslaException:
                response = None

            if response:
                self._site_config[energysite_id].update(response)

        async with self.__update_lock:
            if self._vehicle_list:
                cur_time = round(time.time())
                #  Update the online cars using get_product_list()
                last_update = self._last_attempted_update_time
                _LOGGER.debug(
                    "Get vehicles. Force: %s Time: %s Interval %s",
                    force,
                    cur_time - last_update,
                    ONLINE_INTERVAL,
                )
                if (
                    force
                    or cur_time - last_update >= ONLINE_INTERVAL
                    and update_vehicles
                ):
                    self._product_list = await self.get_product_list()
                    self._vehicle_list = [
                        cars for cars in self._product_list if "vehicle_id" in cars
                    ]
                    for car in self._vehicle_list:
                        self.set_id_vin(car_id=car["id"], vin=car["vin"])
                        self.set_vehicle_id_vin(
                            vehicle_id=car["vehicle_id"], vin=car["vin"]
                        )
                        self.set_car_online(
                            vin=car["vin"], online_status=car["state"] == "online"
                        )
                        self.cars[car["vin"]].update_car_info(car)
                    self._last_attempted_update_time = cur_time

                # Only update online vehicles that haven't been updated recently
                # The throttling is per car's last succesful update
                # Note: This separate check is because there may be individual cars
                # to update.
                car_id = self._update_id(car_id)
                car_vin = self._id_to_vin(car_id)

                for vin, online in self.get_car_online().items():
                    # If specific car_id provided, only update match
                    if (
                        (car_vin and car_vin != vin)
                        or vin not in self.__lock
                        or (vin and self.cars[vin].in_service)
                    ):
                        continue

                    if isinstance(vins, set) and vin not in vins:
                        continue

                    async with self.__lock[vin]:
                        if (
                            (
                                online
                                or (
                                    wake_if_asleep
                                    and self.cars[vin].state in ["asleep", "offline"]
                                )
                            )
                            and (  # pylint: disable=too-many-boolean-expressions
                                self.__update.get(vin)
                            )  # Only update cars with update flag on
                            and (
                                force
                                or vin not in self._last_update_time
                                or (
                                    cur_time - self._last_update_time[vin]
                                    >= self._calculate_next_interval(vin)
                                )
                            )
                        ):
                            tasks.append(_get_and_process_car_data(vin))
                        else:
                            _LOGGER.debug(
                                (
                                    "%s: Skipping update with state %s. Polling: %s. "
                                    "Last update: %s ago. Last parked: %s ago. "
                                    "Last wake up %s ago. "
                                ),
                                vin[-5:],
                                self.cars[vin].state,
                                self.__update.get(vin),
                                cur_time - self._last_update_time[vin],
                                cur_time - self.get_last_park_time(vin=vin),
                                cur_time - self.get_last_wake_up_time(vin=vin),
                            )
            if self._energysite_list and not car_id:
                # do not update energy sites if car_id was a parameter.
                for energysite in self._energysite_list:
                    energysite_id = energysite["energy_site_id"]
                    if (
                        isinstance(energy_site_ids, set)
                        and energysite_id not in energy_site_ids
                    ):
                        continue

                    tasks.append(_get_and_process_site_data(energysite_id))
                    tasks.append(_get_and_process_site_config(energysite_id))

                    if energysite[RESOURCE_TYPE] == RESOURCE_TYPE_BATTERY:
                        tasks.append(_get_and_process_site_summary(energysite_id))

            result = False
            for task in tasks:
                # Update in sequence since establishing a new connection
                # is more expensive because of TLS than the actual update
                # so we want to maximize the chance of reusing the connection
                result |= bool(await task)

            return result

    def get_updates(self, car_id: Text = None, vin: Text = None):
        """Get updates dictionary.

        Parameters
        ----------
        car_id : string
            Identifier for the car on the owner-api endpoint. It is the id
            field for identifying the car across the owner-api endpoint.
            https://tesla-api.timdorr.com/api-basics/vehicles#vehicle_id-vs-id
        vin : string
            VIN number.

        If both car_id and vin is provided. VIN overrides car_id.

        Returns
        -------
        bool or dict of booleans
            If car_id or vin exists, a bool indicating whether updates should be
            processed. Othewise, the entire updates dictionary.

        """
        if car_id and not vin:
            vin = self._id_to_vin(car_id)
        if vin and vin in self.__update:
            return self.__update[vin]
        return self.__update

    def set_updates(
        self, car_id: Text = None, vin: Text = None, value: bool = False
    ) -> None:
        """Set updates dictionary.

        If a vehicle is enabled, the vehicle will force an update on next poll.

        Parameters
        ----------
        car_id : string
            Identifier for the car on the owner-api endpoint. Confusingly it
            is not the vehicle_id field for identifying the car across
            different endpoints.
            https://tesla-api.timdorr.com/api-basics/vehicles#vehicle_id-vs-id
        vin : string
            Vin number
        value : bool
            Whether the specific car_id should be updated.

        Returns
        -------
        None

        """
        if car_id and not vin:
            vin = self._id_to_vin(car_id)
        if vin:
            self.__update[vin] = value
            if self.__update[vin]:
                self.set_last_update_time(vin=vin)
                _LOGGER.debug(
                    "%s: Set Updates enabled; forcing update on next poll by resetting last_update_time",
                    vin[-5:],
                )

    def get_last_update_time(self, car_id: Text = None, vin: Text = None):
        """Get last_update time dictionary.

        Parameters
        ----------
        car_id : string
            Identifier for the car on the owner-api endpoint. It is the id
            field for identifying the car across the owner-api endpoint.
            https://tesla-api.timdorr.com/api-basics/vehicles#vehicle_id-vs-id
            If no car_id, returns the complete dictionary.
        vin : string
            Vin number

        Returns
        -------
        int or dict of ints
            If car_id exists, a int (time.time()) indicating when updates last
            processed. Othewise, the entire updates dictionary.

        """
        if car_id and not vin:
            vin = self._id_to_vin(car_id)
        if vin and vin in self._last_update_time:
            return self._last_update_time[vin]
        return self._last_update_time

    def set_last_update_time(
        self, car_id: Text = None, vin: Text = None, timestamp: float = 0
    ) -> None:
        """Set updated_time for car_id."""
        if car_id and not vin:
            vin = self._id_to_vin(car_id)
        if vin:
            self._last_update_time[vin] = timestamp

    def get_last_park_time(self, car_id: Text = None, vin: Text = None):
        """Get park_time.

        Parameters
        ----------
        car_id : string
            Identifier for the car on the owner-api endpoint. It is the id
            field for identifying the car across the owner-api endpoint.
            https://tesla-api.timdorr.com/api-basics/vehicles#vehicle_id-vs-id
            If no car_id, returns the complete dictionary.
        vin : string
            Vin number

        Returns
        -------
        int or dict of ints
            If car_id exists, a int (time.time()) indicating when car was last
            parked. Othewise, the entire updates dictionary.

        """
        if car_id and not vin:
            vin = self._id_to_vin(car_id)
        if vin and vin in self.__last_parked_timestamp:
            return self.__last_parked_timestamp[vin]
        return self.__last_parked_timestamp

    def set_last_park_time(
        self,
        car_id: Text = None,
        vin: Text = None,
        timestamp: float = 0,
        shift_state: Text = None,
    ) -> None:
        """Set park_time for car_id."""
        if car_id and not vin:
            vin = self._id_to_vin(car_id)
        if vin:
            _LOGGER.debug(
                "%s: Resetting last_parked_timestamp to: %s shift_state %s",
                vin[-5:],
                timestamp,
                shift_state,
            )
            self.__last_parked_timestamp[vin] = timestamp

    def get_last_wake_up_time(self, car_id: Text = None, vin: Text = None):
        """Get wakeup_time.

        Parameters
        ----------
        car_id : string
            Identifier for the car on the owner-api endpoint. It is the id
            field for identifying the car across the owner-api endpoint.
            https://tesla-api.timdorr.com/api-basics/vehicles#vehicle_id-vs-id
            If no car_id, returns the complete dictionary.
        vin : string
            VIN number

        Returns
        -------
        int or dict of ints
            If car_id exists, a int (time.time()) indicating when car was last
            waken up. Othewise, the entire updates dictionary.

        """
        if car_id and not vin:
            vin = self._id_to_vin(car_id)
        if vin and vin in self._last_wake_up_time:
            return self._last_wake_up_time[vin]
        return self._last_wake_up_time

    def set_last_wake_up_time(
        self, car_id: Text = None, vin: Text = None, timestamp: float = 0
    ) -> None:
        """Set wakeup_time for car_id."""
        if car_id and not vin:
            vin = self._id_to_vin(car_id)
        if vin:
            _LOGGER.debug("%s: Resetting last_wake_up_time to: %s", vin[-5:], timestamp)
            self._last_wake_up_time[vin] = timestamp

    def set_car_online(
        self, car_id: Text = None, vin: Text = None, online_status: bool = True
    ) -> None:
        """Set online status for car_id.

        Will also update "last_wake_up_time" if the car changes from offline
        to online

        Parameters
        ----------
        car_id : string
            Identifier for the car on the owner-api endpoint.
        vin : string
            VIN number

        online_status : boolean
            True if the car is online (awake)
            False if the car is offline (out of reach or sleeping)


        """
        if car_id and not vin:
            vin = self._id_to_vin(car_id)
        if vin and self.get_car_online(vin=vin) != online_status:
            _LOGGER.debug(
                "%s: Changing car_online from %s to %s",
                vin[-5:],
                self.get_car_online(vin=vin),
                online_status,
            )
            self.car_online[vin] = online_status
            if online_status:
                self.set_last_wake_up_time(vin=vin, timestamp=round(time.time()))

    def get_car_online(self, car_id: Text = None, vin: Text = None):
        """Get online status for car_id or all cars.

        Parameters
        ----------
        car_id : string
            Identifier for the car on the owner-api endpoint. It is the id
            field for identifying the car across the owner-api endpoint.
            https://tesla-api.timdorr.com/api-basics/vehicles#vehicle_id-vs-id
        vin : string
            VIN number.

        If both car_id and vin is provided. VIN overrides car_id.

        Returns
        -------
        dict or boolean
            If car_id or vin exists, a boolean with the online status for a
            single car.
            Othewise, the entire dictionary with all cars.

        """
        if car_id and not vin:
            vin = self._id_to_vin(car_id)
        if vin and vin in self.car_online:
            return self.car_online[vin]
        return self.car_online

    def is_car_online(self, car_id: Text = None, vin: Text = None) -> bool:
        """Alias for get_car_online for better readability."""
        return self.get_car_online(car_id=car_id, vin=vin)

    def set_id_vin(self, car_id: Text, vin: Text) -> None:
        """Update mappings of car_id <--> vin."""
        car_id = str(car_id)
        self.__id_vin_map[car_id] = vin
        self.__vin_id_map[vin] = car_id

    def set_vehicle_id_vin(self, vehicle_id: Text, vin: Text) -> None:
        """Update mappings of vehicle_id <--> vin."""
        vehicle_id = str(vehicle_id)
        self.__vehicle_id_vin_map[vehicle_id] = vin
        self.__vin_vehicle_id_map[vin] = vehicle_id

    @property
    def update_interval(self) -> int:
        """Return update_interval.

        Returns
            int: The number of seconds between updates

        """
        return self._update_interval

    @update_interval.setter
    def update_interval(self, value: int) -> None:
        """Set update_interval."""
        # Sometimes receive a value of None
        if value and value < 0:
            value = UPDATE_INTERVAL
        if value and value:
            _LOGGER.debug("Update interval set to %s.", value)
            self._update_interval = int(value)

    def set_update_interval_vin(
        self, car_id: Text = None, vin: Text = None, value: int = None
    ) -> None:
        """Set update interval for specific vin."""

        if car_id and not vin:
            vin = self._id_to_vin(car_id)
        if vin is None:
            return
        if value is None or value < 0:
            _LOGGER.debug("%s: Update interval reset to default.", vin[-5:])
            self._update_interval_vin.pop(vin, None)
        else:
            _LOGGER.debug("%s: Update interval set to %s.", vin[-5:], value)
            self._update_interval_vin.update({vin: value})

    def get_update_interval_vin(self, car_id: Text = None, vin: Text = None) -> int:
        """Get update interval for specific vin or default if no vin specific."""
        if car_id and not vin:
            vin = self._id_to_vin(car_id)
        if vin is None or vin == "":
            return self.update_interval

        return self._update_interval_vin.get(vin, self.update_interval)

    def _id_to_vin(self, car_id: Text) -> Optional[Text]:
        """Return vin for a car_id."""
        return self.__id_vin_map.get(str(car_id))

    def _vin_to_id(self, vin: Text) -> Optional[Text]:
        """Return car_id for a vin."""
        return self.__vin_id_map.get(vin)

    def _vehicle_id_to_vin(self, vehicle_id: Text) -> Optional[Text]:
        """Return vin for a vehicle_id."""
        return self.__vehicle_id_vin_map.get(vehicle_id)

    def _vehicle_id_to_id(self, vehicle_id: Text) -> Optional[Text]:
        """Return car_id for a vehicle_id."""
        return self._vin_to_id(self._vehicle_id_to_vin(vehicle_id))

    def vin_to_vehicle_id(self, vin: Text) -> Optional[Text]:
        """Return vehicle_id for a vin."""
        return self.__vin_vehicle_id_map.get(vin)

    def _update_id(self, car_id: Text) -> Optional[Text]:
        """Update the car_id for a vin."""
        new_car_id = self.__vin_id_map.get(self._id_to_vin(car_id))
        if new_car_id:
            car_id = new_car_id
        return car_id

    def _process_websocket_message(self, data):
        if data["msg_type"] == "data:update":
            update_json = {}
            vehicle_id = int(data["tag"])
            vin = self.__vehicle_id_vin_map[vehicle_id]
            # shift_state,speed,power,est_lat,est_lng,est_heading,est_corrected_lat,est_corrected_lng,
            # native_latitude,native_longitude,native_heading,native_type,native_location_supported
            keys = [
                ("timestamp", int),
                ("shift_state", str),
                ("speed", int),
                ("power", int),
                ("est_lat", float),
                ("est_lng", float),
                ("est_heading", int),
                ("est_corrected_lat", float),
                ("est_corrected_lng", float),
                ("native_latitude", float),
                ("native_longitude", float),
                ("native_heading", float),
                ("native_type", str),
                ("native_location_supported", int),
                # ("soc", int),
                # ("elevation", int),
                # ("range", int),
                # ("est_range", int),
                # ("heading", int),
            ]
            values = data["value"].split(",")
            try:
                for num, value in enumerate(values):
                    update_json[keys[num][0]] = keys[num][1](value) if value else None
                _LOGGER.debug("Updating %s with websocket: %s", vin[-5:], update_json)
                self.__driving[vin]["timestamp"] = update_json["timestamp"]
                if (
                    self.cars[vin].shift_state
                    and self.cars[vin].shift_state != update_json["shift_state"]
                    and (
                        update_json["shift_state"] is None
                        or update_json["shift_state"] == "P"
                    )
                ):
                    self.set_last_park_time(
                        vin=vin,
                        timestamp=update_json["timestamp"] / 1000,
                        shift_state=update_json["shift_state"],
                    )
                self.cars[vin].shift_state = update_json["shift_state"]
                self.cars[vin].speed = update_json["speed"]
                self.cars[vin].power = update_json["power"]
                self.cars[vin].latitude = update_json["est_corrected_lat"]
                self.cars[vin].longitude = update_json["est_corrected_lng"]
                self.cars[vin].heading = update_json["est_heading"]
                self.cars[vin].native_latitude = update_json["native_latitude"]
                self.cars[vin].native_longitude = update_json["native_longitude"]
                self.cars[vin].native_heading = update_json["native_heading"]
                self.cars[vin].native_type = update_json["native_type"]
                self.cars[vin].native_location_supported = update_json[
                    "native_location_supported"
                ]

            except ValueError as ex:
                _LOGGER.debug(
                    "Websocket for %s malformed: %s\n%s", vin[-5:], values, ex
                )
        for func in self.__websocket_listeners:
            func(data)

    def _process_websocket_disconnect(self, data):
        vehicle_id = int(data["tag"])
        vin = self.__vehicle_id_vin_map[vehicle_id]
        _LOGGER.debug("Disconnected %s from websocket", vin[-5:])

    def _get_vehicle_ids_for_api(self, path_vars):
        vehicle_id = path_vars.get("vehicle_id")
        if not vehicle_id:
            return None, None

        vehicle_id = str(vehicle_id)
        if vehicle_id in self.__id_vin_map:
            car_id = vehicle_id
            car_vin = self.__id_vin_map.get(vehicle_id)
            return car_id, car_vin

        if vehicle_id in self.__vin_id_map:
            car_id = self.__vin_id_map.get(vehicle_id)
            car_vin = vehicle_id
            return car_id, car_vin

        _LOGGER.error("Could not determine correct vehicle ID for API communication: '%s'", vehicle_id)
        return None, None


    async def api(
        self,
        name: str,
        path_vars=None,
        wake_if_asleep: bool = False,
        **kwargs,
    ):
        """Perform api request for given endpoint name, with keyword arguments as parameters.

        Code from https://github.com/tdorssers/TeslaPy/blob/master/teslapy/__init__.py#L242-L277 under MIT

        Parameters
        ----------
        name : string
            Endpoint command, e.g., STATUS. See https://github.com/zabuldon/teslajsonpy/blob/dev/teslajsonpy/endpoints.json
        path_vars : dict
            Path variables to be replaced. Defaults to None. For vehicle_id reference see https://tesla-api.timdorr.com/api-basics/vehicles#vehicle_id-vs-id
        wake_if_asleep : bool
            If a sleeping vehicle should be woken before making the api call.
        **kwargs :
            Arguments to pass to underlying Tesla command. See https://tesla-api.timdorr.com/vehicle/commands

        Raises
        ------
        ValueError:
            If endpoint name is not found
        NotImplementedError:
            Endpoint method not implemented
        ValueError:
            Path variables missing

        Returns
        -------
        dict
            Tesla json response object.

        """
        path_vars = path_vars or {}
        # use of car_id was deprecated on the new API but VIN can be used on both new and old so always use vin
        car_id, car_vin = self._get_vehicle_ids_for_api(path_vars)
        if path_vars.get("vehicle_id"):
            if car_vin:
                path_vars["vehicle_id"] = car_vin
            else:
                _LOGGER.warning("WARNING: could not set vehicle_id to car_vin, will attempt to send without overriding but this might cause issues!")

        # Load API endpoints once
        if not self.endpoints:
            try:
                data = pkgutil.get_data(__name__, "endpoints.json")
                self.endpoints = orjson.loads(  # pylint: disable=no-member
                    data.decode()
                )
                _LOGGER.debug("%d endpoints loaded", len(self.endpoints))
            except (IOError, ValueError, JSONDecodeError):
                _LOGGER.error("No endpoints loaded")
        # Lookup endpoint name
        try:
            endpoint = self.endpoints[name]
        except KeyError as ex:
            raise ValueError("Unknown endpoint name " + name) from ex
        # Only JSON is supported
        if endpoint.get("CONTENT", "JSON") != "JSON" or name == "STATUS":
            raise NotImplementedError(f"Endpoint {name} not implemented")
        # Substitute path variables in URI
        try:
            uri = endpoint["URI"].format(**path_vars)
        except KeyError as ex:
            raise ValueError(f"{name} requires path variable {ex}") from ex
        method = endpoint["TYPE"].lower()

        # Old @wake_up decorator condensed here
        if wake_if_asleep:
            if not path_vars.get("vehicle_id"):
                raise ValueError(
                    "wake_if_asleep only supported on endpoints with 'vehicle_id' path variable"
                )

            # If we already know the car is asleep, go ahead and wake it
            if not self.is_car_online(car_id=car_id, vin=car_vin):
                await self.wake_up(car_id=car_id)
                return await self.__post_with_retries(
                    "", method=method, data=kwargs, url=uri
                )

            # We think the car is awake, lets try the api call:
            try:
                response = await self.__post_with_retries(
                    "", method=method, data=kwargs, url=uri
                )
            except TeslaException as ex:
                # Don't bother to wake and retry if it's not retryable
                if not ex.retryable:
                    raise ex
                response = None
            # It may fail if the car slept since the last api update
            if not valid_result(response):
                # Assumed it failed because it was asleep and we didn't know it
                await self.wake_up(car_id=car_id)
                response = await self.__post_with_retries(
                    "", method=method, data=kwargs, url=uri
                )
            return response

        # Perform request using given keyword arguments as parameters
        # wake_if_asleep is False so we do not retry if the car is asleep
        # or if the car is unavailable
        return await self.__post_with_retries_except_unavailable(
            "", method=method, data=kwargs, url=uri
        )

    @retry(
        wait=custom_wait,
        retry=custom_retry,
        stop=stop_after_delay(MAX_API_RETRY_TIME),
        reraise=True,
    )
    async def __post_with_retries(self, command, method="post", data=None, url=""):
        """Call connection.post with retries for common exceptions.

        Retries if the car is unavailable.
        """
        return await self.__connection.post(command, method=method, data=data, url=url)

    @retry(
        wait=custom_wait,
        retry=custom_retry_except_unavailable,
        stop=stop_after_delay(MAX_API_RETRY_TIME),
        reraise=True,
    )
    async def __post_with_retries_except_unavailable(
        self, command, method="post", data=None, url=""
    ):
        """Call connection.post with retries for common exceptions.

        Does not retry if the car is unavailable. This should be
        used when wake_if_asleep is False since its unlikely the
        car will suddenly become available if its offline/sleep.
        """
        return await self.__connection.post(command, method=method, data=data, url=url)



#  SPDX-License-Identifier: Apache-2.0
"""
Python Package for controlling Tesla API.

For more details about this api, please refer to the documentation at
https://github.com/zabuldon/teslajsonpy
"""
import asyncio
import base64
import calendar
import datetime
import hashlib
import logging
import secrets
import time
from typing import Dict, Text

from json import JSONDecodeError
import aiohttp
from bs4 import BeautifulSoup
import httpx
import orjson
import yarl
from yarl import URL

from teslajsonpy.const import (
    API_URL,
    CLIENT_ID,
    AUTH_DOMAIN,
    DRIVING_INTERVAL,
    DOMAIN_KEY,
    WEBSOCKET_TIMEOUT,
    WS_URL,
)
from teslajsonpy.exceptions import IncompleteCredentials, TeslaException

_LOGGER = logging.getLogger(__name__)


class Connection:
    """Connection to Tesla Motors API."""

    def __init__(
        self,
        websession: httpx.AsyncClient,
        email: Text = None,
        password: Text = None,
        access_token: Text = None,
        refresh_token: Text = None,
        authorization_token: Text = None,
        expiration: int = 0,
        auth_domain: str = AUTH_DOMAIN,
        client_id: str = CLIENT_ID,
        api_proxy_url: str = None,
    ) -> None:
        """Initialize connection object."""
        self.user_agent: Text = "TeslaApp/4.10.0"
        self.client_id = client_id
        if api_proxy_url is None:
            self.baseurl: Text = DOMAIN_KEY.get(
                auth_domain[auth_domain.rfind(".") :], API_URL
            )
        else:
            self.baseurl: Text = api_proxy_url
        self.websocket_url: Text = WS_URL
        self.api: Text = "/api/1/"
        self.expiration: int = expiration
        self.access_token = access_token
        self.id_token = None
        self.head = None
        self.refresh_token = refresh_token
        self.websession = websession
        self.email = email
        self.password = password
        self.token_refreshed = False
        self.code_verifier: Text = secrets.token_urlsafe(64)
        self.code_challenge = str(
            base64.urlsafe_b64encode(
                hashlib.sha256(self.code_verifier.encode()).hexdigest().encode()
            ),
            "utf-8",
        )
        self.code = authorization_token
        self.sso_oauth: Dict[Text, Text] = {}
        if self.access_token:
            self.__sethead(access_token=self.access_token, expiration=self.expiration)
            _LOGGER.debug("Connecting with existing access token")
        self.websocket = None
        self.mfa_code: Text = ""
        self.auth_domain: URL = URL(auth_domain)

    async def get(self, command):
        """Get data from API."""
        return await self.post(command, "get", None)

    async def post(self, command, method="post", data=None, url=""):
        """Post data to API."""
        now = calendar.timegm(datetime.datetime.now().timetuple())
        _LOGGER.debug(
            "Token expiration in %s",
            str(datetime.timedelta(seconds=self.expiration - now)),
        )
        if now > self.expiration:
            self.token_refreshed = False
            auth = {}
            _LOGGER.debug("Oauth expiration detected")
            if (self.code or (self.email and self.password)) and (
                not self.sso_oauth
                or (
                    now > self.sso_oauth.get("expires_in", 0)
                    and not self.sso_oauth.get("refresh_token")
                )
            ):
                if self.email and self.password:
                    _LOGGER.debug("Getting sso auth code using credentials")
                    self.code = await self.get_authorization_code(
                        self.email, self.password, mfa_code=self.mfa_code
                    )
                else:
                    _LOGGER.debug("Using existing authorization code")
                auth = await self.get_sso_auth_token(self.code)
            elif self.sso_oauth.get("refresh_token") and now > self.sso_oauth.get(
                "expires_in", 0
            ):
                _LOGGER.debug("Refreshing sso auth code")
                auth = await self.refresh_access_token(
                    refresh_token=self.sso_oauth.get("refresh_token")
                )
            elif self.refresh_token:
                auth = await self.refresh_access_token(refresh_token=self.refresh_token)
            if auth and all(
                (
                    auth.get(item)
                    for item in ["access_token", "refresh_token", "expires_in"]
                )
            ):
                self.sso_oauth = {
                    "access_token": auth["access_token"],
                    "refresh_token": auth["refresh_token"],
                    "expires_in": auth["expires_in"] + now,
                }
                self.id_token = auth["id_token"]
                self.refresh_token = auth["refresh_token"]
                _LOGGER.debug("Saved new auth info %s", self.sso_oauth)
            else:
                _LOGGER.debug("Unable to refresh sso oauth token")
                if auth:
                    _LOGGER.debug("Auth returned %s", auth)
                self.code = None
                self.sso_oauth = {}
                raise IncompleteCredentials("Need oauth credentials")
            if auth.get("created_at"):
                # use server time if available
                self.__sethead(
                    access_token=auth["access_token"],
                    expiration=auth["expires_in"] + auth["created_at"],
                )
            else:
                self.__sethead(
                    access_token=auth["access_token"], expires_in=auth["expires_in"]
                )
            self.token_refreshed = True
            _LOGGER.debug("Successfully refreshed oauth")
        if not url:
            url = f"{self.api}{command}"
        return await self.__open(url, method=method, headers=self.head, data=data)

    def __sethead(self, access_token: Text, expires_in: int = 30, expiration: int = 0):
        """Set HTTP header."""
        self.access_token = access_token
        if expiration > 0:
            self.expiration = expiration
        else:
            now = calendar.timegm(datetime.datetime.now().timetuple())
            self.expiration = now + expires_in
        self.head = {
            "Authorization": f"Bearer {access_token}",
            "User-Agent": self.user_agent,
            "X-Tesla-User-Agent": self.user_agent
        }

    async def __open(
        self,
        url: Text,
        method: Text = "get",
        headers=None,
        cookies=None,
        data=None,
        baseurl: Text = "",
    ) -> None:
        """Open url."""
        headers = headers or {}
        cookies = cookies or {}
        if not baseurl:
            baseurl = self.baseurl
        url: URL = URL(baseurl).join(URL(url))
        debug = _LOGGER.isEnabledFor(logging.DEBUG)
        if debug:
            _LOGGER.debug("%s: %s %s", method, url, data)

        try:
            if data:
                if method == "get":
                    resp: httpx.Response = await getattr(self.websession, method)(
                        str(url), params=data, headers=headers, cookies=cookies
                    )
                else:
                    resp: httpx.Response = await getattr(self.websession, method)(
                        str(url), json=data, headers=headers, cookies=cookies
                    )
            else:
                resp: httpx.Response = await getattr(self.websession, method)(
                    str(url), headers=headers, cookies=cookies
                )
            if debug:
                _LOGGER.debug("%s: %s", resp.status_code, resp.text)
            if resp.status_code > 299:
                if resp.status_code == 401:
                    if data and data.get("error") == "invalid_token":
                        raise TeslaException(resp.status_code, "invalid_token")
                elif resp.status_code == 408:
                    raise TeslaException(resp.status_code, "vehicle_unavailable")
                raise TeslaException(resp.status_code)
            data = orjson.loads(resp.content)  # pylint: disable=no-member
            if data.get("error"):
                # known errors:
                #     'vehicle unavailable: {:error=>"vehicle unavailable:"}',
                #     "upstream_timeout", "vehicle is curently in service"
                if debug:
                    _LOGGER.debug(
                        "Raising exception for : %s",
                        f'{data.get("error")}:{data.get("error_description")}',
                    )
                raise TeslaException(
                    f'{data.get("error")}:{data.get("error_description")}'
                )
        except httpx.HTTPStatusError as exception_:
            raise TeslaException(exception_.request.status_code) from exception_
        except JSONDecodeError as exception_:
            raise TeslaException("Error decoding response into json") from exception_
        return data

    async def websocket_connect(self, vin: int, vehicle_id: int, **kwargs):
        """Connect to Tesla streaming websocket.

        Args:
            vin (int): vin of vehicle
            vehicle_id (int): vehicle_id from Tesla api
            on_message (function): function to call on a valid message. It must
                                   process a json delivered in data
            on_disconnect (function): function to call on a disconnect message. It must
                                   process a json delivered in data

        """

        async def _process_messages() -> None:
            """Start Async WebSocket Listener."""
            nonlocal last_message_time
            nonlocal disconnected
            async for msg in self.websocket:
                _LOGGER.debug("msg: %s", msg)
                if msg.type == aiohttp.WSMsgType.BINARY:
                    try:
                        msg_json = orjson.loads(msg.data)  # pylint: disable=no-member
                        if msg_json["msg_type"] == "control:hello":
                            _LOGGER.debug(
                                "%s:Succesfully connected to websocket %s",
                                vin[-5:],
                                self.websocket_url,
                            )
                        if msg_json["msg_type"] == "data:update":
                            last_message_time = time.time()
                        if (
                            msg_json["msg_type"] == "data:error"
                            and msg_json["value"] == "Can't validate token. "
                        ):
                            raise TeslaException(
                                "Can't validate token for websocket connection."
                            )
                        if (
                            msg_json["msg_type"] == "data:error"
                            and msg_json["value"] == "disconnected"
                        ):
                            if kwargs.get("on_disconnect"):
                                kwargs.get("on_disconnect")(msg_json)
                            disconnected = True
                        if kwargs.get("on_message"):
                            kwargs.get("on_message")(msg_json)
                    except JSONDecodeError:
                        _LOGGER.debug("Received bad websocket message: %s", msg.data)
                        break
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    _LOGGER.debug("WSMsgType error")
                    break

        disconnected = False
        last_message_time = time.time()
        timeout = last_message_time + DRIVING_INTERVAL
        if not self.websocket or self.websocket.closed:
            _LOGGER.debug("%s:Connecting to websocket %s", vin[-5:], self.websocket_url)
            self.websocket = await self.websession.ws_connect(self.websocket_url)
            loop = asyncio.get_event_loop()
            loop.create_task(_process_messages())
        while not (
            disconnected
            or time.time() - last_message_time > WEBSOCKET_TIMEOUT
            or time.time() > timeout
        ):
            _LOGGER.debug("%s:Trying to subscribe to websocket", vin[-5:])
            await self.websocket.send_json(
                data={
                    "msg_type": "data:subscribe_oauth",
                    "token": self.access_token,
                    "value": "shift_state,speed,power,est_lat,est_lng,est_heading,est_corrected_lat,est_corrected_lng,native_latitude,native_longitude,native_heading,native_type,native_location_supported",
                    # "value": "speed,odometer,soc,elevation,est_heading,est_lat,est_lng,power,shift_state,range,est_range,heading",
                    # old values
                    "tag": f"{vehicle_id}",
                    "created:timestamp": round(time.time() * 1000),
                }
            )
            await asyncio.sleep(WEBSOCKET_TIMEOUT - 1)
        _LOGGER.debug(
            "%s:Exiting websocket_connect",
            vin[-5:],
        )

    # async def websocket_connect2(self, vin: int, vehicle_id: int, **kwargs):
    #     """Connect to Tesla streaming websocket.

    #     Args:
    #         vin (int): vin of vehicle
    #         vehicle_id (int): vehicle_id from Tesla api
    #         on_message (function): function to call on a valid message. It must
    #                                process a json delivered in data
    #         on_disconnect (function): function to call on a disconnect message. It must
    #                                process a json delivered in data

    #     """

    #     async def _process_messages() -> None:
    #         """Start Async WebSocket Listener."""
    #         async for msg in self.websocket[vin]["websocket"]:
    #             _LOGGER.debug("%s:msg: %s", vin[-5:], msg)
    #             if msg.type == aiohttp.WSMsgType.BINARY:
    #                 msg_json = orjson.loads(msg.data)
    #                 if msg_json["msg_type"] == "control:hello":
    #                     _LOGGER.debug(
    #                         "%s:Succesfully connected to websocket %s on %s",
    #                         vin[-5:],
    #                         self.websocket_url,
    #                         task,
    #                     )
    #                 if (
    #                     msg_json["msg_type"] == "data:error"
    #                     and msg_json["value"] == "Can't validate token. "
    #                 ):
    #                     raise TeslaException(
    #                         "Can't validate token for websocket connection."
    #                     )
    #                 if (
    #                     msg_json["msg_type"] == "data:error"
    #                     and msg_json["value"] == "disconnected"
    #                 ):
    #                     if self.websocket[vin].kwargs.get("on_disconnect"):
    #                         self.websocket[vin].kwargs.get("on_disconnect")()
    #                     self.websocket[vin].pop(None)
    #                     _LOGGER.debug(
    #                         "%s:Disconnecting from websocket on %s", vin[-5:], task
    #                     )
    #                 await self.websocket[vin]["websocket"].close()
    #                 if kwargs.get("on_message"):
    #                     kwargs.get("on_message")(msg_json)
    #             elif msg.type == aiohttp.WSMsgType.ERROR:
    #                 _LOGGER.debug("WSMsgType error")
    #                 break

    #     self.websocket.setdefault(vin, {"websocket": None, "kwargs": kwargs})
    #     if (
    #         not self.websocket[vin]["websocket"]
    #         or self.websocket[vin]["websocket"].closed
    #     ):
    #         _LOGGER.debug("%s:Connecting to websocket %s", vin[-5:], self.websocket_url)
    #         self.websocket[vin]["websocket"] = await self.websession.ws_connect(
    #             self.websocket_url
    #         )
    #         loop = asyncio.get_event_loop()
    #         task = loop.create_task(_process_messages())
    #     _LOGGER.debug(
    #         "%s:Trying to subscribe to websocket: %s", vin[-5:], self.access_token
    #     )

    #     await self.websocket[vin]["websocket"].send_json(
    #         data={
    #             "msg_type": "data:subscribe_oauth",
    #             # "token": "self.access_token",
    #             "token": self.access_token,
    #             "value": "speed,odometer,soc,elevation,est_heading,est_lat,est_lng,power,shift_state,range,est_range,heading",
    #             "tag": f"{vehicle_id}",
    #         }
    #     )

    async def close(self) -> None:
        """Close connection."""
        await self.websession.aclose()
        _LOGGER.debug("Connection closed.")

    async def get_authorization_code(
        self,
        email: Text,
        password: Text,
        mfa_code: Text = "",
        mfa_device: int = 0,
        retry_limit: int = 3,
    ) -> Text:
        """Get authorization code from the oauth3 login method."""
        # https://tesla-api.timdorr.com/api-basics/authentication#step-2-obtain-an-authorization-code
        # pylint: disable=too-many-locals
        if not (email and password):
            _LOGGER.debug("No email or password for login; unable to login.")
            return
        url = self.get_authorization_code_link(new=True)
        resp = await self.websession.get(str(url.update_query({"login_hint": email})))
        html = resp.text
        if resp.history:
            for item in resp.history:
                if (
                    item.status_code in [301, 302, 303, 304, 305, 306, 307, 308]
                    and resp.url.host != self.auth_domain.host
                ):
                    _LOGGER.debug(
                        "Detected %s redirect from %s to %s; changing proxy host",
                        item.status_code,
                        item.url.host,
                        resp.url.host,
                    )
                    self.auth_domain = self.auth_domain.with_host(str(resp.url.host))
                    url = self.get_authorization_code_link()
        soup: BeautifulSoup = BeautifulSoup(html, "html.parser")
        data = get_inputs(soup)
        data["identity"] = email
        data["credential"] = password
        transaction_id: Text = data.get("transaction_id")
        for attempt in range(retry_limit):
            _LOGGER.debug("Attempt #%s", attempt)
            resp = await self.websession.post(str(url), data=data)
            _process_resp(resp)
            if not resp.history:
                html = resp.text
                if "/mfa/verify" in html:
                    _LOGGER.debug("Detected MFA request")
                    mfa_resp = await self.websession.get(
                        str(
                            self.auth_domain.with_path(
                                "/oauth2/v3/authorize/mfa/factors"
                            )
                        ),
                        params={"transaction_id": transaction_id},
                    )
                    _process_resp(mfa_resp)
                    # {
                    #     "data": [
                    #         {
                    #         "dispatchRequired": false,
                    #         "id": "X-4Y-44e4-b9a4-54e114a13c40",
                    #         "name": "Pixel",
                    #         "factorType": "token:software",
                    #         "factorProvider": "TESLA",
                    #         "securityLevel": 1,
                    #         "activatedAt": "2021-02-10T23:53:40.000Z",
                    #         "updatedAt": "2021-02-10T23:54:20.000Z"
                    #         }
                    #     ]
                    # }
                    mfa_json = orjson.loads(mfa_resp.text)  # pylint: disable=no-member
                    if len(mfa_json.get("data", [])) >= 1:
                        factor_id = mfa_json["data"][mfa_device]["id"]
                    if not mfa_code:
                        _LOGGER.debug("No MFA provided")
                        _LOGGER.debug("MFA Devices: %s", mfa_json["data"])
                        raise IncompleteCredentials(
                            "MFA Code missing", devices=mfa_json["data"]
                        )
                    mfa_resp = await self.websession.post(
                        str(
                            self.auth_domain.with_path(
                                "/oauth2/v3/authorize/mfa/verify"
                            )
                        ),
                        json={
                            "transaction_id": transaction_id,
                            "factor_id": factor_id,
                            "passcode": mfa_code,
                        },
                    )
                    _process_resp(mfa_resp)
                    mfa_json = orjson.loads(mfa_resp.text)  # pylint: disable=no-member
                    if not (
                        mfa_json["data"].get("approved")
                        and mfa_json["data"].get("valid")
                    ):
                        _LOGGER.debug("MFA Code invalid")
                        raise IncompleteCredentials(
                            "MFA Code invalid", devices=mfa_json["data"]
                        )
                    resp = await self.websession.post(str(url), data=data)
                    _process_resp(resp)
            await asyncio.sleep(3)
        if not resp.history or not URL(str(resp.history[-1].url)).query.get("code"):
            _LOGGER.debug("Failed to authenticate")
            raise IncompleteCredentials("Unable to login with credentials")
        code_url = URL(str(resp.history[-1].url))
        _LOGGER.debug("Found code %s", code_url.query.get("code"))
        return code_url.query.get("code")

    def get_authorization_code_link(self, new=False) -> yarl.URL:
        """Get authorization code url for the oauth3 login method."""
        # https://tesla-api.timdorr.com/api-basics/authentication#step-2-obtain-an-authorization-code
        if new:
            self.code_verifier: Text = secrets.token_urlsafe(64)
            self.code_challenge = str(
                base64.urlsafe_b64encode(
                    hashlib.sha256(self.code_verifier.encode()).hexdigest().encode()
                ),
                "utf-8",
            )
        state = secrets.token_urlsafe(64)
        query = {
            "client_id": "ownerapi",
            "code_challenge": self.code_challenge,
            "code_challenge_method": "S256",
            "redirect_uri": "https://auth.tesla.com/void/callback",
            "response_type": "code",
            "scope": "openid email offline_access",
            "state": state,
        }
        url = self.auth_domain.with_path("/oauth2/v3/authorize")
        url = url.update_query(query)
        return url

    async def get_sso_auth_token(self, code):
        """Get sso auth token."""
        # https://tesla-api.timdorr.com/api-basics/authentication#step-2-obtain-an-authorization-code
        _LOGGER.debug("Requesting new sso oauth token using sso auth code")
        if not code:
            _LOGGER.debug("No authorization code provided")
            return
        oauth = {
            "client_id": "ownerapi",
            "grant_type": "authorization_code",
            "code": code,
            "code_verifier": self.code_verifier,
            "redirect_uri": "https://auth.tesla.com/void/callback",
        }
        auth = await self.websession.post(
            str(self.auth_domain.with_path("/oauth2/v3/token")),
            data=oauth,
        )
        try:
            return orjson.loads(auth.text)  # pylint: disable=no-member
        except JSONDecodeError:
            _LOGGER.debug("Received bad auth response: %s", auth.text)
            return None

    async def refresh_access_token(self, refresh_token):
        """Refresh access token from sso."""
        # https://tesla-api.timdorr.com/api-basics/authentication#refreshing-an-access-token
        if not refresh_token:
            _LOGGER.debug("Missing refresh token")
            return
        _LOGGER.debug("Refreshing access token with refresh_token")
        oauth = {
            "client_id": self.client_id,
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "scope": "openid email offline_access",
        }
        auth = await self.websession.post(
            str(self.auth_domain.with_path("/oauth2/v3/token")),
            data=oauth,
        )
        try:
            return orjson.loads(auth.text)  # pylint: disable=no-member
        except JSONDecodeError:
            _LOGGER.debug("Received bad auth response: %s", auth.text)
            return None

    async def get_bearer_token(self, access_token):
        """Get bearer token. This is used by the owners API."""
        # This appears deprecated as of March 21, 2022: https://github.com/timdorr/tesla-api/issues/548
        # https://tesla-api.timdorr.com/api-basics/authentication#step-4-exchange-bearer-token-for-access-token
        if not access_token:
            _LOGGER.debug("Missing access token")
            return
        _LOGGER.debug("Exchanging bearer token with access token:")
        oauth = {
            "client_id": self.client_id,
            "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
        }
        head = {
            "Authorization": f"Bearer {access_token}",
        }
        auth = await self.websession.post(
            f"{self.baseurl}/oauth/token", headers=head, data=oauth
        )
        try:
            return orjson.loads(auth.text)  # pylint: disable=no-member
        except JSONDecodeError:
            _LOGGER.debug("Received bad auth response: %s", auth.text)
            return None


def get_inputs(soup: BeautifulSoup, searchfield=None) -> Dict[str, str]:
    """Parse soup for form with searchfield."""
    searchfield = searchfield or {"id": "form"}
    data = {}
    form = soup.find("form", searchfield)
    if not form:
        form = soup.find("form")
    if form:
        for field in form.find_all("input"):
            try:
                data[field["name"]] = ""
                if field["type"] and field["type"] == "hidden":
                    data[field["name"]] = field["value"]
            except BaseException:  # pylint: disable=broad-except
                pass
    return data


def _process_resp(resp) -> Text:
    if resp.history:
        for item in resp.history:
            _LOGGER.debug("%s: redirected from\n%s", item.request.method, item.url)
    url = str(resp.request.url)
    method = resp.request.method
    status = resp.status_code
    reason = resp.reason_phrase
    headers = resp.request.headers
    _LOGGER.debug(
        "%s: \n%s with\n%s\n returned %s:%s with response %s",
        method,
        url,
        headers,
        status,
        reason,
        resp.headers,
    )
    return url
